"""Test NFTHolderCondition contract."""
from contracts_lib_py.conditions.transfer_nft import TransferNFTCondition


def test_transfer_did_condition_contract():
    transfer_nft_condition = TransferNFTCondition('TransferNFTCondition')
    assert transfer_nft_condition
    assert isinstance(transfer_nft_condition, TransferNFTCondition), \
        f'{transfer_nft_condition} is not instance of TransferNFTCondition'



from contracts_lib_py.conditions.condition_base import ConditionBase


class HashLockCondition(ConditionBase):
    """Class representing the HashLockCondition contract."""
    CONTRACT_NAME = 'HashLockCondition'

    def fulfill(self, agreement_id, preimage, account):
        """

        :param agreement_id: id of the agreement, hex str
        :param preimage:
        :param account: Account instance
        :return:
        """
        return self._fulfill(
            agreement_id,
            preimage,
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )

    def hash_values(self, preimage):
        """
        Hash the values of the hash lock condition.

        :param preimage:
        :return:
        """
        return self._hash_values(preimage)



import os
import pathlib

from web3 import Web3

from contracts_lib_py import Keeper
from contracts_lib_py.utils import get_account
from contracts_lib_py.web3_provider import Web3Provider
from contracts_lib_py.keeper import Keeper
from contracts_lib_py.contract_handler import ContractHandler
from contracts_lib_py.web3.http_provider import CustomHTTPProvider



PUBLISHER_INDEX = 1
CONSUMER_INDEX = 0


def get_keeper_url():
    if os.getenv('KEEPER_URL'):
        return os.getenv('KEEPER_URL')
    return 'http://localhost:8545'


def setup_keeper():
    Web3Provider._web3 = Web3(CustomHTTPProvider(get_keeper_url()))
    ContractHandler.artifacts_path = os.path.expanduser('~/.nevermined/nevermined-contracts/artifacts')
    Keeper.get_instance(artifacts_path=ContractHandler.artifacts_path)


def get_network_name():
    return Keeper.get_instance().get_network_name(Keeper.get_instance().get_network_id())


def get_resource_path(dir_name, file_name):
    base = os.path.realpath(__file__).split(os.path.sep)[1:-1]
    if dir_name:
        return pathlib.Path(os.path.join(os.path.sep, *base, dir_name, file_name))
    else:
        return pathlib.Path(os.path.join(os.path.sep, *base, file_name))


def init_ocn_tokens(ocn, account, amount=100):
    ocn.accounts.request_tokens(account, amount)
    Keeper.get_instance().token.token_approve(
        Keeper.get_instance().dispenser.address,
        amount,
        account
    )


def get_publisher_account():
    return get_account(0)


def get_consumer_account():
    return get_account(0)



import logging
import os
import time
from collections import namedtuple

from eth_keys import KeyAPI
from eth_utils import big_endian_to_int
from web3 import Web3
from web3.contract import ContractEvent
from web3._utils.encoding import to_bytes
from web3._utils.threads import Timeout

from contracts_lib_py.account import Account
from contracts_lib_py.web3_provider import Web3Provider

Signature = namedtuple('Signature', ('v', 'r', 's'))

logger = logging.getLogger(__name__)


def generate_multi_value_hash(types, values):
    """
    Return the hash of the given list of values.
    This is equivalent to packing and hashing values in a solidity smart contract
    hence the use of `solidityKeccak`.

    :param types: list of solidity types expressed as strings
    :param values: list of values matching the `types` list
    :return: bytes
    """
    assert len(types) == len(values)
    return Web3.solidityKeccak(
        types,
        values
    )


def prepare_prefixed_hash(msg_hash):
    """

    :param msg_hash:
    :return:
    """
    return generate_multi_value_hash(
        ['string', 'bytes32'],
        ["\x19Ethereum Signed Message:\n32", msg_hash]
    )


def add_ethereum_prefix_and_hash_msg(text):
    """
    This method of adding the ethereum prefix seems to be used in web3.sign/ecRecover.

    :param text: str any str to be signed / used in recovering address from a signature
    :return: hash of prefixed text according to the recommended ethereum prefix
    """
    prefixed_msg = f"\x19Ethereum Signed Message:\n{len(text)}{text}"
    return Web3.keccak(text=prefixed_msg)


def get_public_key_from_address(web3, account):
    """

    :param web3:
    :param account:
    :return:
    """
    _hash = web3.keccak(text='verify signature.')
    signature = web3.sign(_hash, account.address, account.password)
    signature = split_signature(web3, to_bytes(hexstr=signature))
    signature_vrs = Signature(signature.v % 27,
                              big_endian_to_int(signature.r),
                              big_endian_to_int(signature.s))
    prefixed_hash = prepare_prefixed_hash(_hash)
    pub_key = KeyAPI.PublicKey.recover_from_msg_hash(prefixed_hash,
                                                     KeyAPI.Signature(vrs=signature_vrs))
    assert pub_key.to_checksum_address() == account.address, \
        'recovered address does not match signing address.'
    return pub_key


def to_32byte_hex(web3, val):
    """

    :param web3:
    :param val:
    :return:
    """
    return web3.toBytes(val).rjust(32, b'\0')


def convert_to_bytes(web3, data):
    """

    :param web3:
    :param data:
    :return:
    """
    return web3.toBytes(text=data)


def convert_to_string(web3, data):
    """

    :param web3:
    :param data:
    :return:
    """
    return web3.toHex(data)


def convert_to_text(web3, data):
    """

    :param web3:
    :param data:
    :return:
    """
    return web3.toText(data)


def split_signature(web3, signature):
    """

    :param web3:
    :param signature: signed agreement hash, hex str
    :return:
    """
    assert len(signature) == 65, f'invalid signature, ' \
                                 f'expecting bytes of length 65, got {len(signature)}'
    v = web3.toInt(signature[-1])
    r = to_32byte_hex(web3, int.from_bytes(signature[:32], 'big'))
    s = to_32byte_hex(web3, int.from_bytes(signature[32:64], 'big'))
    if v != 27 and v != 28:
        v = 27 + v % 2

    return Signature(v, r, s)


def get_account(index):
    name = 'PARITY_ADDRESS' if not index else f'PARITY_ADDRESS{index}'
    pswrd_name = 'PARITY_PASSWORD' if not index else f'PARITY_PASSWORD{index}'
    keyfile_name = 'PARITY_KEYFILE' if not index else f'PARITY_KEYFILE{index}'

    address = os.getenv(name)
    if not address:
        return None

    pswrd = os.getenv(pswrd_name)
    if not pswrd:
        pswrd = os.getenv(f'{address}_PASSWORD')
    key_file = os.getenv(keyfile_name)
    if not key_file:
        key_file = os.getenv(f'{address}_KEYFILE')

    return Account(Web3.toChecksumAddress(address), pswrd, key_file)


def process_tx_receipt(tx_hash, event_instance, event_name, agreement_id=None):
    """
    Wait until the tx receipt is processed.

    :param tx_hash: hash of the transaction
    :param event_instance: instance of ContractEvent
    :param event_name: name of the event to subscribe, str
    :param agreement_id: hex str
    :return:
    """
    if not isinstance(tx_hash, bytes):
        raise TypeError(f'first argument should be bytes type, '
                        f'got type {type(tx_hash)} and value {tx_hash}')
    if not isinstance(event_instance, ContractEvent):
        raise TypeError(f'second argument should be a ContractEvent, '
                        f'got {event_instance} of type {type(event_instance)}')
    web3 = Web3Provider.get_web3()
    try:
        web3.eth.waitForTransactionReceipt(tx_hash, timeout=20)
    except Timeout:
        logger.info(f'Waiting for {event_name} transaction receipt timed out. '
                    f'Cannot verify receipt and event.')
        return False

    receipt = web3.eth.getTransactionReceipt(tx_hash)
    event_logs = event_instance.processReceipt(receipt) if receipt else None
    if event_logs:
        logger.info(f'Success: got {event_name} event after fulfilling condition.')
        logger.debug(
            f'Success: got {event_name} event after fulfilling condition. {receipt}, '
            f'::: {event_logs}')
    else:
        logger.debug(f'Something is not right, cannot find the {event_name} event after calling the'
                     f' fulfillment condition. This is the transaction receipt {receipt}')

    if receipt and receipt.status == 0:
        logger.warning(
            f'Transaction failed: tx_hash {tx_hash.hex()}, tx event {event_name}, receipt '
            f'{receipt}, id {agreement_id}')
        return False

    return True


def process_fulfill_condition(args, condition_contract, condition_id, logger_handle, keeper,
                              num_tries=10):
    if not logger_handle:
        logger_handle = logger

    contract_name = condition_contract.CONTRACT_NAME
    agreement_id = args[0]
    for i in range(num_tries):
        try:
            tx_hash = condition_contract.fulfill(
                *args
            )
            success = process_tx_receipt(
                tx_hash,
                getattr(condition_contract.contract.events, condition_contract.FULFILLED_EVENT)(),
                f'{contract_name}.Fulfilled'
            )
            if success or keeper.condition_manager.get_condition_state(condition_id) == 2:
                logger_handle.info(f'Done {contract_name}.fulfill for agreement {agreement_id}')
                break

            logger_handle.debug(
                f'done trial {i} {contract_name}.fulfill for agreement {agreement_id}, success?: '
                f'{bool(success)}')
            time.sleep(2)

        except Exception as e:
            if keeper.condition_manager.get_condition_state(condition_id) == 2:
                logger_handle.info(f'Done {contract_name}.fulfill for agreement {agreement_id}')
                break

            logger_handle.debug(f'{contract_name}.fulfill error {agreement_id}: {e}', exc_info=1)
            if i == (num_tries - 1):
                logger_handle.debug(
                    f'{contract_name}.fulfill for agreement {agreement_id} FAILED with error: {e}')
            else:
                logger_handle.debug(
                    f'Error when doing {contract_name}.fulfill for agreement {agreement_id}: retrying trial # {i}')
                time.sleep(2)



from contracts_lib_py.templates.template_manager import TemplateStoreManager

def test_template():
    template_store_manager = TemplateStoreManager('TemplateStoreManager')
    assert template_store_manager.get_num_templates() == 15



from web3 import HTTPProvider

from contracts_lib_py.web3.request import make_post_request


class CustomHTTPProvider(HTTPProvider):
    """
    Override requests to control the connection pool to make it blocking.
    """

    def make_request(self, method, params):
        self.logger.debug("Making request HTTP. URI: %s, Method: %s",
                          self.endpoint_uri, method)
        request_data = self.encode_rpc_request(method, params)
        raw_response = make_post_request(
            self.endpoint_uri,
            request_data,
            **self.get_request_kwargs()
        )
        response = self.decode_rpc_response(raw_response)
        self.logger.debug("Getting response HTTP. URI: %s, "
                          "Method: %s, Response: %s",
                          self.endpoint_uri, method, response)
        return response



from contracts_lib_py.conditions.condition_base import ConditionBase

class AccessProofCondition(ConditionBase):
    """Class representing the AccessProofCondition contract."""
    CONTRACT_NAME = 'AccessProofCondition'

    def fulfill(self, agreement_id, hash, buyer, provider, cipher, proof, account):
        """
        Fulfill the access condition.

        :param agreement_id: id of the agreement, hex str
        :param document_id: refers to the DID in which secret store will issue the decryption
        keys, DID
        :param grantee_address: is the address of the granted user, str
        :param account: Account instance
        :return: true if the condition was successfully fulfilled, bool
        """
        return self._fulfill(
            agreement_id,
            int(hash, 16),
            [int(buyer[0], 16), int(buyer[1], 16)],
            [int(provider[0], 16), int(provider[1], 16)],
            [int(cipher[0], 16), int(cipher[1], 16)],
            bytes.fromhex(proof[2:]),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'gas': 2000000,
                      'keyfile': account.key_file}
        )

    def hash_values(self, hash, buyer, provider):
        """
        Hast the values of the document_id with the grantee address.

        :param document_id: refers to the DID in which secret store will issue the decryption
        keys, DID
        :param grantee_address: is the address of the granted user, str
        :return: hex str
        """
        res = self._hash_values(int(hash, 16), [int(buyer[0], 16), int(buyer[1], 16)], [int(provider[0], 16), int(provider[1], 16)])
        return res




"""
This is copied from Web3 python library to control the `requests`
session parameters.
"""
import lru
import requests
from requests.adapters import HTTPAdapter

from web3._utils.caching import (
    generate_cache_key,
)


def _remove_session(key, session):
    session.close()


_session_cache = lru.LRU(8, callback=_remove_session)


def _get_session(*args, **kwargs):
    cache_key = generate_cache_key((args, kwargs))
    if cache_key not in _session_cache:
        # This is the main change from original Web3 `_get_session`
        session = requests.sessions.Session()
        session.mount('http://', HTTPAdapter(pool_connections=25, pool_maxsize=25, pool_block=True))
        session.mount('https://',
                      HTTPAdapter(pool_connections=25, pool_maxsize=25, pool_block=True))
        _session_cache[cache_key] = session
    return _session_cache[cache_key]


def make_post_request(endpoint_uri, data, *args, **kwargs):
    kwargs.setdefault('timeout', 10)
    session = _get_session(endpoint_uri)
    response = session.post(endpoint_uri, data=data, *args, **kwargs)
    response.raise_for_status()

    return response.content



import logging

from contracts_lib_py.templates.template_base import TemplateBase

logger = logging.getLogger('nftSalesTemplate')


class NFTSalesTemplate(TemplateBase):
    """Class representing the NFTSalesTemplate contract."""
    CONTRACT_NAME = 'NFTSalesTemplate'



import json
import pathlib

import pytest
from web3 import HTTPProvider, Web3

from tests.resources.helper_functions import (
    get_consumer_account,
    get_publisher_account,
    setup_keeper,
    get_keeper_url
)


@pytest.fixture(autouse=True)
def setup_all(request):
    if 'integration' in request.keywords:
        return
    setup_keeper()


@pytest.fixture
def publisher_account():
    return get_publisher_account()


@pytest.fixture
def consumer_account():
    return get_consumer_account()


@pytest.fixture
def web3_instance():
    return Web3(HTTPProvider(get_keeper_url()))

@pytest.fixture
def external_contract():
    name = 'ERC721'
    address = '0x18bdFAf7Cc2B66a4Cfa7e069693CD1a9B639A69b'

    with pathlib.Path(__file__).parent / 'resources/data/ERC721.json' as p:
        abi = json.loads(p.read_text())['abi']
        return (address, abi, name)


"""Test DIDRegistry contract."""
import secrets
import uuid

import pytest
from eth_utils import add_0x_prefix
from web3 import Web3

from contracts_lib_py import Keeper
from contracts_lib_py.contract_handler import ContractHandler
from contracts_lib_py.didregistry import DIDRegistry
from contracts_lib_py.nft_upgradeable import NFTUpgradeable
from tests.resources.helper_functions import get_consumer_account, get_publisher_account


def new_did():
    return uuid.uuid4().hex + uuid.uuid4().hex


def test_did_registry_contract():
    did_registry = DIDRegistry.get_instance()
    assert did_registry
    assert isinstance(did_registry, DIDRegistry)


def test_did_registry_get_block_number_updated():
    did_registry = DIDRegistry.get_instance()
    test_id = secrets.token_hex(32)
    assert did_registry.get_block_number_updated(test_id) == 0


def test_register():
    register_account = get_publisher_account()
    did_registry = DIDRegistry.get_instance()
    w3 = Web3
    did_seed = new_did()
    asset_id = did_registry.hash_did(did_seed, register_account.address)

    checksum_test = w3.keccak(text='checksum')
    value_test = 'http://localhost:5000'
    # register DID-> URL
    assert did_registry.register(
        did_seed, checksum_test, url=value_test, account=register_account
    ) is True

    with pytest.raises(Exception):
        did_registry.get_did_owner('did:nv' + asset_id)

    assert did_registry.get_did_owner(asset_id) == register_account.address


def test_register_with_invalid_params():
    register_account = get_publisher_account()
    did_registry = DIDRegistry.get_instance()
    w3 = Web3

    did_seed = new_did()
    asset_id = did_registry.hash_did(did_seed, register_account.address)
    did_test = 'did:nv:' + asset_id

    checksum_test = w3.keccak(text='checksum')
    value_test = 'http://localhost:5000'
    # No checksum provided
    with pytest.raises(ValueError):
        did_registry.register(did_seed, '', url=value_test, account=None)
    # Invalid checksum  provided
    with pytest.raises(ValueError):
        did_registry.register(did_seed, did_test, url=value_test, account=None)

    # No account provided
    with pytest.raises(ValueError):
        did_registry.register(did_seed, checksum_test, url=value_test, account=None)


def test_providers():
    register_account = get_publisher_account()
    consumer_account = get_consumer_account()
    did_registry = DIDRegistry.get_instance()

    w3 = Web3
    did_seed = new_did()
    asset_id = did_registry.hash_did(did_seed, register_account.address)
    # asset_id = add_0x_prefix(asset_id)

    checksum_test = w3.keccak(text='checksum')
    value_test = 'http://localhost:5000'
    test_address = w3.toChecksumAddress('068ed00cf0441e4829d9784fcbe7b9e26d4bd8d0')
    # register DID-> URL
    with pytest.raises(AssertionError):
        did_registry.register(
            did_seed, checksum_test, url=value_test, account=register_account,
            providers=consumer_account.address
        )
    did_registry.register(
        did_seed, checksum_test, url=value_test, account=register_account, providers=[test_address]
    )
    unknown_asset_id = add_0x_prefix(new_did())
    providers = did_registry.get_did_providers(unknown_asset_id)
    assert len(providers) == 0

    assert did_registry.is_did_provider(asset_id, register_account.address) is False

    providers = did_registry.get_did_providers(asset_id)
    assert len(providers) == 1 and providers[0] == test_address
    assert did_registry.is_did_provider(asset_id, test_address) is True

    removed = did_registry.remove_provider(asset_id, test_address, register_account)
    assert removed
    providers = did_registry.get_did_providers(asset_id)
    assert len(providers) == 0
    assert did_registry.is_did_provider(asset_id, test_address) is False

    did_registry.add_provider(asset_id, test_address, register_account)
    providers = did_registry.get_did_providers(asset_id)
    assert len(providers) == 1 and providers[0] == test_address
    assert did_registry.is_did_provider(asset_id, test_address) is True


def test_transfer_ownership():
    register_account = get_publisher_account()

    did_registry = DIDRegistry.get_instance()
    w3 = Web3

    did_seed = new_did()
    asset_id = did_registry.hash_did(did_seed, register_account.address)

    checksum_test = w3.keccak(text='checksum')
    value_test = 'http://localhost:5000'
    test_address = w3.toChecksumAddress('068ed00cf0441e4829d9784fcbe7b9e26d4bd8d0')

    did_registry.register(did_seed, checksum_test, url=value_test, account=register_account,
                          providers=[test_address])

    first_owner = did_registry.get_did_owner(asset_id)
    assert first_owner == register_account.address

    did_registry.transfer_did_ownership(asset_id, test_address, register_account)
    assert did_registry.get_did_owner(asset_id) == test_address


def test_grant_permissions():
    register_account = get_publisher_account()

    did_registry = DIDRegistry.get_instance()
    w3 = Web3

    did_seed = new_did()
    asset_id = did_registry.hash_did(did_seed, register_account.address)

    checksum_test = w3.keccak(text='checksum')
    value_test = 'http://localhost:5000'
    test_address = w3.toChecksumAddress('068ed00cf0441e4829d9784fcbe7b9e26d4bd8d0')

    did_registry.register(did_seed, checksum_test, url=value_test, account=register_account,
                          providers=[test_address])

    assert not did_registry.get_permission(asset_id, test_address)

    did_registry.grant_permission(asset_id, test_address, register_account)

    assert did_registry.get_permission(asset_id, test_address)

    did_registry.revoke_permission(asset_id, test_address, register_account)

    assert not did_registry.get_permission(asset_id, test_address)


def test_provenance_events():
    register_account = get_publisher_account()
    did_registry = DIDRegistry.get_instance()
    w3 = Web3

    did_seed = new_did()
    asset_id = did_registry.hash_did(did_seed, register_account.address)

    checksum_test = w3.keccak(text='checksum')
    activity_id = new_did()
    provenance_id = new_did()
    url = 'http://localhost:5000'
    test_address = w3.toChecksumAddress('068ed00cf0441e4829d9784fcbe7b9e26d4bd8d0')

    assert did_registry.register(did_seed, checksum_test, url=url, account=register_account,
                                 providers=[test_address], activity_id=activity_id) is True

    did_registry.used(provenance_id, asset_id, test_address, activity_id, "",
                      account=register_account, attributes="used test")

    assert len(did_registry.get_did_provenance_events(Web3.toBytes(hexstr=asset_id))) == 2


def test_provenance_from_registry():
    register_account = get_publisher_account()
    did_registry = DIDRegistry.get_instance()
    w3 = Web3

    did_seed = new_did()
    asset_id = did_registry.hash_did(did_seed, register_account.address)

    checksum_test = w3.keccak(text='checksum')
    activity_id = new_did()
    provenance_id = new_did()
    url = 'http://localhost:5000'
    test_address = w3.toChecksumAddress('068ed00cf0441e4829d9784fcbe7b9e26d4bd8d0')

    assert did_registry.register(did_seed, checksum_test, url=url, account=register_account,
                                 providers=[test_address], activity_id=activity_id) is True

    assert register_account.address == did_registry.get_provenance_owner(asset_id)

    did_registry.used(provenance_id, asset_id, test_address, activity_id, "",
                      account=register_account, attributes="used test")

    provenance_entry = did_registry.get_provenance_entry(provenance_id)
    assert ("0x" + activity_id, provenance_entry['activity_id'])
    assert ("0x" + asset_id, provenance_entry['did'])
    assert (3, provenance_entry['method'])
    assert (register_account.address, provenance_entry['created_by'])


def test_delegate_provenance():
    register_account = get_publisher_account()
    delegated_account = get_consumer_account()
    did_registry = DIDRegistry.get_instance()
    w3 = Web3

    did_seed = new_did()
    asset_id = did_registry.hash_did(did_seed, register_account.address)

    checksum_test = w3.keccak(text='checksum')
    activity_id = new_did()
    provenance_id = new_did()
    url = 'http://localhost:5000'
    test_address = w3.toChecksumAddress('068ed00cf0441e4829d9784fcbe7b9e26d4bd8d0')

    assert did_registry.register(did_seed, checksum_test, url=url, account=register_account,
                                 providers=[test_address], activity_id=activity_id) is True

    assert did_registry.is_provenance_delegate(Web3.toBytes(hexstr=asset_id),
                                               register_account.address) is False
    assert did_registry.is_provenance_delegate(Web3.toBytes(hexstr=asset_id),
                                               delegated_account.address) is False
    assert did_registry.add_did_provenance_delegate(Web3.toBytes(hexstr=asset_id),
                                                    delegated_account.address,
                                                    register_account) is True
    assert did_registry.is_provenance_delegate(Web3.toBytes(hexstr=asset_id),
                                               delegated_account.address) is True
    assert did_registry.remove_did_provenance_delegate(Web3.toBytes(hexstr=asset_id),
                                                       delegated_account.address,
                                                       register_account) is True
    assert did_registry.is_provenance_delegate(Web3.toBytes(hexstr=asset_id),
                                               delegated_account.address) is False


def test_search_multiple_provenance_event_tests():
    register_account = get_publisher_account()
    did_registry = DIDRegistry.get_instance()
    w3 = Web3

    did_seed = new_did()
    asset_id = did_registry.hash_did(did_seed, register_account.address)

    checksum_test = w3.keccak(text='checksum')
    activity_id = new_did()
    provenance_id = new_did()
    derived_did = new_did()
    url = 'http://localhost:5000'
    test_address = w3.toChecksumAddress('068ed00cf0441e4829d9784fcbe7b9e26d4bd8d0')

    assert did_registry.register(did_seed, checksum_test, url=url, account=register_account,
                                 providers=[test_address], activity_id=activity_id) is True

    did_registry.used(provenance_id, asset_id, register_account.address, activity_id, "",
                      account=register_account, attributes="used test")

    did_registry.was_derived_from(new_did(), derived_did, asset_id, register_account.address,
                                  activity_id,
                                  account=register_account, attributes="was derived from")

    did_registry.was_associated_with(new_did(), asset_id, register_account.address, activity_id,
                                     account=register_account, attributes="was associated with")

    did_registry.acted_on_behalf(new_did(), asset_id, register_account.address,
                                 register_account.address, activity_id, '',
                                 account=register_account, attributes="acted on behalf")

    assert len(did_registry.get_provenance_method_events('WAS_GENERATED_BY',
                                                         Web3.toBytes(hexstr=asset_id))) == 1
    assert len(
        did_registry.get_provenance_method_events('USED', Web3.toBytes(hexstr=asset_id))) == 1
    assert len(did_registry.get_provenance_method_events('WAS_DERIVED_FROM',
                                                         Web3.toBytes(hexstr=derived_did))) == 1
    assert len(did_registry.get_provenance_method_events('WAS_ASSOCIATED_WITH',
                                                         Web3.toBytes(hexstr=asset_id))) == 1

    assert len(did_registry.get_provenance_method_events('ACTED_ON_BEHALF',
                                                         Web3.toBytes(hexstr=asset_id))) == 1


def test_royalties_are_valid():
    register_account = get_publisher_account()
    did_registry = DIDRegistry.get_instance()
    w3 = Web3

    did_seed = new_did()
    asset_id = did_registry.hash_did(did_seed, register_account.address)

    checksum_test = w3.keccak(text='checksum')
    activity_id = new_did()
    new_owner_address = w3.toChecksumAddress('068ed00cf0441e4829d9784fcbe7b9e26d4bd8d0')
    someone_address = w3.toChecksumAddress('0x00a329c0648769A73afAc7F9381E08FB43dBEA72')
    url = 'http://localhost:5000'
    royalties = 20
    cap = 0

    assert did_registry.register_mintable_did(did_seed, checksum_test, url, cap, royalties, account=register_account,
                                              providers=[someone_address], activity_id=activity_id) is True

    did_registry.transfer_did_ownership(asset_id, new_owner_address, register_account)

    assert did_registry.are_royalties_valid(asset_id, [80, 20], [someone_address, register_account.address]) is True
    assert did_registry.are_royalties_valid(asset_id, [90], [someone_address]) is False
    assert did_registry.are_royalties_valid(asset_id, [90, 10], [someone_address, register_account.address]) is False


def test_get_did():
    register_account = get_publisher_account()
    did_registry = DIDRegistry.get_instance()
    w3 = Web3

    did_seed = new_did()
    asset_id = did_registry.hash_did(did_seed, register_account.address)

    checksum_test = w3.keccak(text='checksum')
    activity_id = new_did()
    someone_address = w3.toChecksumAddress('0x00a329c0648769A73afAc7F9381E08FB43dBEA72')
    url = 'http://localhost:5000'
    royalties = 20
    cap = 0

    assert did_registry.register_mintable_did(did_seed, checksum_test, url, cap, royalties, account=register_account,
                                              providers=[someone_address], activity_id=activity_id) is True

    registered_did = did_registry.get_did_register(asset_id)
    assert registered_did['owner'] == register_account.address
    assert registered_did['url'] == url
    assert registered_did['lastUpdatedBy'] == register_account.address
    assert registered_did['blockNumberUpdated'] > 0
    assert len(registered_did['providers']) == 1
    assert registered_did['nftSupply'] == 0
    assert registered_did['mintCap'] == cap
    assert registered_did['royalties'] == royalties


def test_nft():
    register_account = get_publisher_account()
    did_registry = DIDRegistry.get_instance()
    nft_upgradeable = NFTUpgradeable.get_instance()
    w3 = Web3

    did_seed = new_did()
    asset_id = did_registry.hash_did(did_seed, register_account.address)

    someone_address = "0x00a329c0648769A73afAc7F9381E08FB43dBEA72"
    checksum_test = w3.keccak(text='checksum')
    activity_id = new_did()
    cap = 10
    royalties = 0
    url = 'http://localhost:5000'
    test_address = w3.toChecksumAddress('068ed00cf0441e4829d9784fcbe7b9e26d4bd8d0')

    assert did_registry.register_mintable_did(did_seed, checksum_test, url, cap, royalties, account=register_account,
                                              providers=[test_address], activity_id=activity_id) is True

    balance = nft_upgradeable.balance(register_account.address, asset_id)
    assert balance == 0
    balance_consumer = nft_upgradeable.balance(someone_address, asset_id)
    assert balance_consumer == 0

    did_registry.mint(asset_id, 10, account=register_account)
    assert balance + 10 == nft_upgradeable.balance(register_account.address, asset_id)
    assert nft_upgradeable.transfer_nft(asset_id, someone_address, 1, register_account)
    assert nft_upgradeable.balance(register_account.address, asset_id) == 9
    assert nft_upgradeable.balance(someone_address, asset_id) == balance_consumer + 1
    did_registry.burn(asset_id, 9, account=register_account)
    assert balance == nft_upgradeable.balance(register_account.address, asset_id)


def test_nft_approval():
    did_registry = DIDRegistry.get_instance()
    w3 = Web3
    keeper = Keeper(ContractHandler.artifacts_path)

    assert did_registry.is_nft_approved_for_all(w3.toChecksumAddress('068ed00cf0441e4829d9784fcbe7b9e26d4bd8d0'),
                                                keeper.transfer_nft_condition.address) is True

    assert did_registry.is_nft_approved_for_all(w3.toChecksumAddress('068ed00cf0441e4829d9784fcbe7b9e26d4bd8d0'),
                                                keeper.did_sales_template.address) is False






import logging
import os

from eth_utils import big_endian_to_int
import nevermined_contracts

from contracts_lib_py.agreements.agreement_manager import AgreementStoreManager
from contracts_lib_py.conditions.access import AccessCondition
from contracts_lib_py.conditions.access_proof import AccessProofCondition
from contracts_lib_py.conditions.compute_execution import ComputeExecutionCondition
from contracts_lib_py.conditions.condition_manager import ConditionStoreManager
from contracts_lib_py.conditions.escrow_reward import EscrowPaymentCondition
from contracts_lib_py.conditions.nft_escrow_reward import NFTEscrowPaymentCondition
from contracts_lib_py.conditions.nft721_escrow_reward import NFT721EscrowPaymentCondition
from contracts_lib_py.conditions.hash_lock import HashLockCondition
from contracts_lib_py.conditions.lock_reward import LockPaymentCondition
from contracts_lib_py.conditions.nft_access import NFTAccessCondition
from contracts_lib_py.conditions.nft_holder import NFTHolderCondition
from contracts_lib_py.conditions.nft_lock_condition import NFTLockCondition
from contracts_lib_py.conditions.sign import SignCondition
from contracts_lib_py.conditions.threshold import ThresholdCondition
from contracts_lib_py.conditions.transfer_did import TransferDIDOwnershipCondition
from contracts_lib_py.conditions.transfer_nft import TransferNFTCondition
from contracts_lib_py.conditions.white_listing import WhitelistingCondition
from contracts_lib_py.didregistry import DIDRegistry
from contracts_lib_py.dispenser import Dispenser
from contracts_lib_py.generic_contract import GenericContract, GenericContractExternal
from contracts_lib_py.nft_upgradeable import NFTUpgradeable
from contracts_lib_py.templates.access_template import AccessTemplate
from contracts_lib_py.templates.access_proof_template import AccessProofTemplate
from contracts_lib_py.templates.nft_access_proof_template import NFTAccessProofTemplate
from contracts_lib_py.templates.nft_access_swap_template import NFTAccessSwapTemplate
from contracts_lib_py.templates.nft_sales_with_access_template import NFTSalesWithAccessTemplate
from contracts_lib_py.templates.compute_execution_template import EscrowComputeExecutionTemplate
from contracts_lib_py.templates.did_sales_template import DIDSalesTemplate
from contracts_lib_py.templates.nft_access_template import NFTAccessTemplate
from contracts_lib_py.templates.nft_sales_template import NFTSalesTemplate
from contracts_lib_py.templates.template_manager import TemplateStoreManager
from contracts_lib_py.token import Token
from contracts_lib_py.utils import (add_ethereum_prefix_and_hash_msg, generate_multi_value_hash,
                                    split_signature)
from contracts_lib_py.wallet import Wallet
from contracts_lib_py.web3.signature import SignatureFix
from contracts_lib_py.web3_provider import Web3Provider


class Keeper(object):
    """The Keeper class aggregates all contracts in the Nevermined Protocol node."""

    DEFAULT_NETWORK_NAME = 'development'
    _network_name_map = {
        1: 'mainnet',
        2: 'Morden',
        3: 'Ropsten',
        4: 'rinkeby',
        42: 'Kovan',
        77: 'POA_Sokol',
        99: 'POA_Core',
        137: 'matic',
        1337: 'geth-localnet',
        8996: 'spree',
        8997: 'polygon-localnet',
        42220: 'celo',
        44787: 'celo-alfajores',
        62320: 'celo-baklava',
        80001: 'mumbai',
        1313161554: 'aurora',
        1313161555: 'aurora-testnet',
        1313161556: 'aurora-betanet'
    }

    def __init__(self, artifacts_path=None, contract_names=None, external_contracts=[]):
        self.network_name = Keeper.get_network_name(Keeper.get_network_id())
        self.artifacts_path = artifacts_path or nevermined_contracts.get_artifacts_path()

        # This api is not provided by polygon
        # Not sure if it makes sense to use `eth.accounts` since in order to
        # interact with the accounts in the node we need to be running our own node
        try:
            self.accounts = Web3Provider.get_web3().eth.accounts
        except ValueError as e:
            # method does not exist or is not available
            if e.args[0]['code'] == -32601:
                self.accounts = []
            else:
                raise

        self._contract_name_to_instance = {}
        self._external_contract_name_to_instance = {}
        if contract_names:
            for name in contract_names:
                contract = GenericContract(name)
                self._contract_name_to_instance = contract
                setattr(self, name, contract)

        for (address, abi, name) in external_contracts:
            contract = GenericContractExternal(address, abi, name)
            self._external_contract_name_to_instance[name] = contract
            setattr(self, name, contract)

        # optional contracts (contracts that may not exist in certain networks)
        self.access_proof_template = self._try_optional_contract(AccessProofTemplate)
        self.access_proof_condition = self._try_optional_contract(AccessProofCondition)
        self.dispenser = self._try_optional_contract(Dispenser)
        self.token = self._try_optional_contract(Token)
        self.nft_upgradeable = self._try_optional_contract(NFTUpgradeable)
        self.nft_escrow_payment_condition = self._try_optional_contract(NFTEscrowPaymentCondition)
        self.nft721_escrow_payment_condition = self._try_optional_contract(NFT721EscrowPaymentCondition)
        self.nft_access_proof_template = self._try_optional_contract(NFTAccessProofTemplate)
        self.nft_access_swap_template = self._try_optional_contract(NFTAccessSwapTemplate)
        self.nft_sales_with_access_template = self._try_optional_contract(NFTSalesWithAccessTemplate)

        # required contracts
        self.did_registry = DIDRegistry.get_instance()
        self.template_manager = TemplateStoreManager.get_instance()
        self.access_template = AccessTemplate.get_instance()
        self.escrow_compute_execution_template = EscrowComputeExecutionTemplate.get_instance()
        self.agreement_manager = AgreementStoreManager.get_instance()
        self.condition_manager = ConditionStoreManager.get_instance()
        self.sign_condition = SignCondition.get_instance()
        self.lock_payment_condition = LockPaymentCondition.get_instance()
        self.escrow_payment_condition = EscrowPaymentCondition.get_instance()
        self.access_condition = AccessCondition.get_instance()
        self.nft_access_condition = NFTAccessCondition.get_instance()
        self.compute_execution_condition = ComputeExecutionCondition.get_instance()
        self.hash_lock_condition = HashLockCondition.get_instance()
        self.threshold_condition = ThresholdCondition.get_instance()
        self.white_listing_condition = WhitelistingCondition.get_instance()
        self.nft_holder_condition = NFTHolderCondition.get_instance()
        self.nft_lock_condition = NFTLockCondition.get_instance()
        self.transfer_nft_condition = TransferNFTCondition.get_instance()
        self.transfer_did_condition = TransferDIDOwnershipCondition.get_instance()
        self.nft_access_template = NFTAccessTemplate.get_instance()
        self.did_sales_template = DIDSalesTemplate.get_instance()
        self.nft_sales_template = NFTSalesTemplate.get_instance()
        templates = {}
        for i in [self.access_template, self.nft_access_template, self.nft_access_proof_template, self.nft_access_swap_template,self.nft_sales_with_access_template, self.did_sales_template, self.nft_sales_template, self.access_proof_template]:
            if i != None:
                templates[i.address] = i
        self.agreement_manager.set_templates(templates)
        contracts = [
            self.token,
            self.did_registry,
            self.template_manager,
            self.access_template,
            self.access_proof_template,
            self.escrow_compute_execution_template,
            self.agreement_manager,
            self.condition_manager,
            self.sign_condition,
            self.lock_payment_condition,
            self.escrow_payment_condition,
            self.nft_escrow_payment_condition,
            self.nft721_escrow_payment_condition,
            self.access_condition,
            self.access_proof_condition,
            self.nft_access_condition,
            self.compute_execution_condition,
            self.hash_lock_condition,
            self.threshold_condition,
            self.white_listing_condition,
            self.nft_holder_condition,
            self.nft_lock_condition,
            self.transfer_nft_condition,
            self.transfer_did_condition,
            self.nft_access_template,
            self.did_sales_template,
            self.nft_access_proof_template,
            self.nft_access_swap_template,
            self.nft_sales_with_access_template,
            self.nft_sales_template
        ]
        if self.dispenser:
            contracts.append(self.dispenser)
        self._contract_name_to_instance = {contract.name: contract
                                           for contract in contracts if contract}

    @staticmethod
    def get_instance(artifacts_path=None, contract_names=None, external_contracts=[]):
        """Return the Keeper instance (singleton)."""
        return Keeper(artifacts_path, contract_names, external_contracts)

    @staticmethod
    def get_network_name(network_id):
        """
        Return the keeper network name based on the current ethereum network id.
        Return `development` for every network id that is not mapped.

        :param network_id: Network id, int
        :return: Network name, str
        """
        if os.environ.get('KEEPER_NETWORK_NAME'):
            logging.debug('keeper network name overridden by an environment variable: {}'.format(
                os.environ.get('KEEPER_NETWORK_NAME')))
            return os.environ.get('KEEPER_NETWORK_NAME')

        return Keeper._network_name_map.get(network_id, Keeper.DEFAULT_NETWORK_NAME)

    @staticmethod
    def get_network_id():
        """
        Return the ethereum network id calling the `web3.version.network` method.

        :return: Network id, int
        """
        return int(Web3Provider.get_web3().net.version)

    @staticmethod
    def sign_hash(msg_hash, account):
        """
        This method use `personal_sign`for signing a message. This will always prepend the
        `\x19Ethereum Signed Message:\n32` prefix before signing.

        :param msg_hash:
        :param account: Account
        :return: signature
        """
        wallet = Wallet(Web3Provider.get_web3(), account.key_file, account.password,
                        account.address)
        s = wallet.sign(msg_hash)
        return s.signature.hex()

    @staticmethod
    def ec_recover(message, signed_message):
        """
        This method does not prepend the message with the prefix `\x19Ethereum Signed Message:\n32`.
        The caller should add the prefix to the msg/hash before calling this if the signature was
        produced for an ethereum-prefixed message.

        :param message:
        :param signed_message:
        :return:
        """
        w3 = Web3Provider.get_web3()
        v, r, s = split_signature(w3, w3.toBytes(hexstr=signed_message))
        signature_object = SignatureFix(vrs=(v, big_endian_to_int(r), big_endian_to_int(s)))
        return w3.eth.account.recoverHash(message, signature=signature_object.to_hex_v_hacked())

    @staticmethod
    def personal_ec_recover(message, signed_message):
        prefixed_hash = add_ethereum_prefix_and_hash_msg(message)
        return Keeper.ec_recover(prefixed_hash, signed_message)

    @staticmethod
    def unlock_account(account):
        """
        Unlock the account.

        :param account: Account
        :return:
        """
        return Web3Provider.get_web3().unlockAccount(account.address, account.password)

    @staticmethod
    def get_ether_balance(address):
        """
        Get balance of an ethereum address.

        :param address: address, bytes32
        :return: balance, int
        """
        return Web3Provider.get_web3().eth.getBalance(address, block_identifier='latest')

    @property
    def contract_name_to_instance(self):
        return self._contract_name_to_instance

    @property
    def external_contract_name_to_instance(self):
        return self._external_contract_name_to_instance

    def get_contract(self, contract_name):
        contract = self.contract_name_to_instance.get(contract_name)
        if contract:
            return contract

        contract = self.external_contract_name_to_instance.get(contract_name)
        if contract:
            return contract

        try:
            return GenericContract(contract_name)
        except Exception as e:
            logging.error(f'Cannot load contract {contract_name}: {e}')
            return None

    @staticmethod
    def generate_multi_value_hash(types, values):
        return generate_multi_value_hash(types, values)

    @staticmethod
    def _try_optional_contract(contract):
        try:
            return contract.get_instance()
        except FileNotFoundError:
            return None



import logging

from contracts_lib_py.templates.template_base import TemplateBase

logger = logging.getLogger('escrowComputeExecutionTemplate')


class EscrowComputeExecutionTemplate(TemplateBase):
    """Class representing the EscrowComputeExecutionTemplate contract."""
    CONTRACT_NAME = 'EscrowComputeExecutionTemplate'



"""Test Dispenser contract."""
import pytest

from contracts_lib_py.dispenser import Dispenser
from tests.resources.helper_functions import get_consumer_account


@pytest.fixture()
def dispenser():
    return Dispenser.get_instance()


def test_dispenser_contract(dispenser):
    assert dispenser
    assert isinstance(dispenser, Dispenser), f'{dispenser} is not instance of Market'


def test_request_tokens(dispenser):
    account = get_consumer_account()
    assert dispenser.request_tokens(100, account), f'{account.address} do not get 100 tokens.'



"""Test PaymentConditions contract."""
from contracts_lib_py.conditions.lock_reward import LockPaymentCondition


def test_lock_payment_condition_contract():
    lock_payment_condition = LockPaymentCondition('LockPaymentCondition')
    assert lock_payment_condition
    assert isinstance(lock_payment_condition, LockPaymentCondition), \
        f'{lock_payment_condition} is not instance of LockPaymentCondition'



"""Test the Account object."""
from contracts_lib_py.account import Account


def test_create_account():
    account = Account('0x213123123', 'pass')
    assert isinstance(account, Account)
    assert account.address == '0x213123123'
    assert account.password == 'pass'



import logging

from contracts_lib_py.templates.template_base import TemplateBase

logger = logging.getLogger('AccessTemplate')


class AccessTemplate(TemplateBase):
    """Class representing the AccessTemplate contract."""
    CONTRACT_NAME = 'AccessTemplate'



from web3 import Web3

from contracts_lib_py.utils import split_signature


def test_split_signature():
    signature = b'\x19\x15!\xecwnX1o/\xdeho\x9a9\xdd9^\xbb\x8c2z\x88!\x95\xdc=\xe6\xafc' \
                b'\x0f\xe9\x14\x12\xc6\xde\x0b\n\xa6\x11\xc0\x1cvv\x9f\x99O8\x15\xf6f' \
                b'\xe7\xab\xea\x982Ds\x0bX\xd9\x94\xa42\x01'
    _signature = split_signature(Web3, signature=signature)
    assert _signature.v == 28
    assert _signature.r == b'\x19\x15!\xecwnX1o/\xdeho\x9a9\xdd9^\xbb\x8c2z\x88!\x95' \
                                b'\xdc=\xe6\xafc\x0f\xe9'
    assert _signature.s == b'\x14\x12\xc6\xde\x0b\n\xa6\x11\xc0\x1cvv\x9f\x99O8\x15' \
                                b'\xf6f\xe7\xab\xea\x982Ds\x0bX\xd9\x94\xa42'





import logging

from web3 import Web3

from contracts_lib_py import ContractBase, utils
from contracts_lib_py.event_filter import EventFilter

logger = logging.getLogger('accessTemplate')


class ConditionBase(ContractBase):
    """Base class for all the Condition contract objects."""
    FULFILLED_EVENT = 'Fulfilled'
    ZERO_ADDRESS = '0x0000000000000000000000000000000000000000'

    def generate_id(self, agreement_id, types, values):
        """
        Generate id for the condition.

        :param agreement_id: id of the agreement, hex str
        :param types: list of types
        :param values: list of values
        :return: id, str
        """
        values_hash = utils.generate_multi_value_hash(types, values)
        return utils.generate_multi_value_hash(
            ['bytes32', 'uint256', 'bytes32'],
            [agreement_id, int(self.address, 16), values_hash.hex()]
        )

    def _fulfill(self, *args, method='fulfill', **kwargs):
        """
        Fulfill the condition.

        :param args:
        :param kwargs:
        :return: true if the condition was successfully fulfilled, bool
        """
        tx_hash = self.send_transaction(method, args, **kwargs)
        if self.is_tx_successful(tx_hash):
            logger.info('Condition fulfilled successfully')
        else:
            logger.error('It looks that there was a problem fulfilling the tx.')
        return tx_hash

    def abort_by_timeout(self, condition_id):
        """

        :param condition_id: id of the condition, str
        :return:
        """
        tx_hash = self.contract.caller.abortByTimeOut(condition_id)
        return self.is_tx_successful(tx_hash)

    def _hash_values(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        return self.contract.caller.hashValues(*args, **kwargs)

    def _hash_values_multi(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        return self.contract.caller.hashValuesMulti(*args, **kwargs)

    def _hash_values_marked(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        return self.contract.caller.hashValuesMarked(*args, **kwargs)

    def subscribe_condition_fulfilled(self, agreement_id, timeout, callback, args,
                                      timeout_callback=None, wait=False,
                                      from_block=None, to_block=None):
        """
        Subscribe to the condition fullfilled event.

        :param agreement_id: id of the agreement, hex str
        :param timeout:
        :param callback:
        :param args:
        :param timeout_callback:
        :param wait: if true block the listener until get the event, bool
        :param from_block: int or None
        :param to_block: int or None
        :return:
        """
        logger.debug(
            f'Subscribing {self.FULFILLED_EVENT} event with agreement id {agreement_id}.')
        return self.subscribe_to_event(
            self.FULFILLED_EVENT,
            timeout,
            {'_agreementId': Web3.toBytes(hexstr=agreement_id)},
            callback=callback,
            timeout_callback=timeout_callback,
            args=args,
            wait=wait,
            from_block=from_block,
            to_block=to_block
        )

    def get_event_filter_for_fulfilled(self, agreement_id=None, from_block='latest',
                                       to_block='latest'):
        _filter = {}
        if agreement_id:
            assert isinstance(agreement_id, str) or isinstance(agreement_id, bytes)
            if isinstance(agreement_id, str) and agreement_id.startswith('0x'):
                agreement_id = Web3.toBytes(hexstr=agreement_id)
            _filter['_agreementId'] = agreement_id

        event_filter = EventFilter(
            self.FULFILLED_EVENT,
            getattr(self.events, self.FULFILLED_EVENT),
            _filter,
            from_block=from_block,
            to_block=to_block
        )
        event_filter.set_poll_interval(0.5)
        return event_filter



import logging

from web3 import Web3
from web3._utils.threads import Timeout

from contracts_lib_py.contract_base import ContractBase
from contracts_lib_py.event_filter import EventFilter
from contracts_lib_py.exceptions import InvalidTransaction
from contracts_lib_py.web3_provider import Web3Provider


class Dispenser(ContractBase):
    """Class representing the Dispenser contract."""

    CONTRACT_NAME = 'Dispenser'

    @staticmethod
    def get_scale():
        return 10 ** 18

    def request_vodkas(self, amount, account):
        self.request_tokens(max(1, int(amount / self.get_scale())), account)

    def request_tokens(self, amount, account):
        """
        Request an amount of tokens for a particular address.
        This transaction has gas cost

        :param amount: Amount of tokens, int
        :param account: Account instance
        :raise InvalidTransaction: Transaction failed
        :return: bool
        """
        address = account.address
        try:
            tx_hash = self.send_transaction(
                'requestTokens',
                (amount,),
                transact={'from': address,
                          'passphrase': account.password,
                          'keyfile': account.key_file}
            )
            logging.debug(f'{address} requests {amount} tokens, returning receipt')
            try:
                receipt = Web3Provider.get_web3().eth.waitForTransactionReceipt(
                    tx_hash, timeout=10)
                logging.debug(f'requestTokens receipt: {receipt}')
            except Timeout:
                receipt = None

            if not receipt:
                return False

            if receipt.status == 0:
                logging.warning(f'request tokens failed: Tx-receipt={receipt}')
                logging.warning(f'request tokens failed: account {address}')
                return False

            # check for emitted events:
            rfe = EventFilter(
                'RequestFrequencyExceeded',
                self.events.RequestFrequencyExceeded,
                argument_filters={'requester': Web3.toBytes(hexstr=address)},
                from_block='latest',
                to_block='latest',
            )
            logs = rfe.get_all_entries(max_tries=5)
            if logs:
                logging.warning(f'request tokens failed RequestFrequencyExceeded')
                logging.info(f'RequestFrequencyExceeded event logs: {logs}')
                return False

            rle = EventFilter(
                'RequestLimitExceeded',
                self.events.RequestLimitExceeded,
                argument_filters={'requester': Web3.toBytes(hexstr=address)},
                from_block='latest',
                to_block='latest',
            )
            logs = rle.get_all_entries(max_tries=5)
            if logs:
                logging.warning(f'request tokens failed RequestLimitExceeded')
                logging.info(f'RequestLimitExceeded event logs: {logs}')
                return False

            return True

        except ValueError as err:
            raise InvalidTransaction(
                f'Requesting {amount} tokens'
                f' to {address} failed with error: {err}'
            )



from contracts_lib_py.conditions.condition_base import ConditionBase


class WhitelistingCondition(ConditionBase):
    """Class representing the WhitelistingCondition contract."""
    CONTRACT_NAME = 'WhitelistingCondition'

    def fulfill(self, agreement_id, list_contract_address, address, from_account):
        """
        Fulfill check whether address is whitelisted in order to fulfill the condition.

        :param agreement_id: id of the agreement, hex str
        :param list_contract_address: List contract address, str
        :param address: address in the list, str
        :param from_account: Account instance
        :return: true if the condition was successfully fulfilled, bool
        """
        return self._fulfill(
            agreement_id,
            list_contract_address,
            address,
            transact={'from': from_account.address,
                      'passphrase': from_account.password,
                      'keyfile': from_account.key_file}
        )

    def hash_values(self, list_contract_address, address):
        """
        HashValues generates the hash of condition inputs with the following parameters

        :param list_contract_address: List contract address, str
        :param address: address in the list, str
        :return: hex str
        """
        return self._hash_values(list_contract_address, address)



from contracts_lib_py.conditions.condition_base import ConditionBase


class NFT721EscrowPaymentCondition(ConditionBase):
    """Class representing the EscrowPaymentCondition contract."""
    CONTRACT_NAME = 'NFT721EscrowPaymentCondition'

    def fulfill(self,
                agreement_id,
                did,
                amount,
                receiver,
                return_address,
                sender_address,
                token_address,
                lock_condition_id,
                release_condition_id,
                account):
        """
        Fulfill the escrow payment condition.

        :param agreement_id: id of the agreement, hex str
        :param did: id of the asset, hex str
        :param amounts: Array of amount of tokens to distribute, int[]
        :param receivers: Array of ethereum address of the receivers, hex str[]
        :param sender_address: ethereum address of the sender, hex str
        :param token_address: the token address to use. If empty  it will use 0x0 address that means ETH payment
        :param lock_condition_id: id of the condition, str
        :param release_condition_id: id of the condition, str
        :param account: Account instance
        :return:
        """
        return self._fulfill(
            agreement_id,
            did,
            amount,
            receiver,
            return_address,
            sender_address,
            self.validate_token_address(token_address),
            lock_condition_id,
            release_condition_id,
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )

    def hash_values(self, did, amount, receiver, return_address, sender_address, token_address, lock_condition_id,
                    release_condition_id):
        """
        Hash the values of the escrow reward condition.

        :param did: id of the asset, hex str
        :param amounts: Array of amount of tokens to distribute, int[]
        :param receivers: Array of ethereum address of the receivers, hex str[]
        :param sender_address: ethereum address of the sender, hex str
        :param token_address: the token address to use. If empty  it will use 0x0 address that means ETH payment
        :param lock_condition_id: id of the condition, str
        :param release_condition_id: id of the condition, str
        :return: hex str
        """
        return self._hash_values(
            did,
            amount,
            receiver,
            return_address,
            sender_address,
            self.validate_token_address(token_address),
            lock_condition_id,
            release_condition_id
        )




"""
    Keeper Contract Base

    All keeper contract inherit from this base class
"""
import logging

from web3 import Web3
from web3._utils.threads import Timeout

from contracts_lib_py.web3.contract import CustomContractFunction
from contracts_lib_py.web3_provider import Web3Provider

logger = logging.getLogger('keeper')


class ContractBase(object):
    """Base class for all contract objects."""
    CONTRACT_NAME = None
    ZERO_ADDRESS = '0x0000000000000000000000000000000000000000'

    def __init__(self, contract_name, dependencies=None):

        assert contract_name, 'param contract_name is required and must ' \
                              'match a valid keeper contract.'
        self.name = contract_name

        if not dependencies or 'ContractHandler' not in dependencies:
            dependencies = dict()
            from contracts_lib_py.contract_handler import ContractHandler
            dependencies['ContractHandler'] = ContractHandler

        self.contract = dependencies['ContractHandler'].get(contract_name)
        self.version = dependencies['ContractHandler'].get_contract_version(contract_name)

    @classmethod
    def get_instance(cls, dependencies=None):
        """
        Return an instance for a contract name.

        :param dependencies:
        :return: Contract base instance
        """
        assert cls is not ContractBase, 'ContractBase is not meant to be used directly.'
        assert cls.CONTRACT_NAME, 'CONTRACT_NAME must be set to a valid keeper contract name.'
        return cls(cls.CONTRACT_NAME, dependencies)

    @property
    def _contract(self):
        return self.contract

    @property
    def address(self):
        """Return the ethereum address of the solidity contract deployed
        in current keeper network.
        """
        return self._contract.address

    @property
    def events(self):
        """Expose the underlying contract's events.

        :return:
        """
        return self.contract.events

    @staticmethod
    def to_checksum_address(address):
        """
        Validate the address provided.

        :param address: Address, hex str
        :return: address, hex str
        """
        return Web3.toChecksumAddress(address)

    @staticmethod
    def to_checksum_addresses(addresses):
        """
        Calculate the checksum of an addresses array

        :param address: Address, hex str[]
        :return: address, hex str[]
        """

        hash = []
        for address in addresses:
            hash.append(Web3.toChecksumAddress(address))
        return hash

    @staticmethod
    def validate_token_address(token_address):
        if token_address is None or not Web3.isAddress(token_address):
            return ContractBase.ZERO_ADDRESS
        return token_address

    @staticmethod
    def get_tx_receipt(tx_hash):
        """
        Get the receipt of a tx.

        :param tx_hash: hash of the transaction
        :return: Tx receipt
        """
        try:
            tx_receipt = Web3Provider.get_web3().eth.waitForTransactionReceipt(tx_hash, timeout=20)
        except Timeout:
            logger.info('Waiting for transaction receipt timed out.')
            return
        except ValueError as e:
            logger.error(f'Waiting for transaction receipt failed: {e}')
            return

        return tx_receipt

    def is_tx_successful(self, tx_hash):
        receipt = self.get_tx_receipt(tx_hash)
        return bool(receipt and receipt.status == 1)

    def subscribe_to_event(self, event_name, timeout, event_filter, callback=None,
                           timeout_callback=None, args=None, wait=False,
                           from_block='latest', to_block='latest'):
        """
        Create a listener for the event choose.

        :param event_name: name of the event to subscribe, str
        :param timeout:
        :param event_filter:
        :param callback:
        :param timeout_callback:
        :param args:
        :param wait: if true block the listener until get the event, bool
        :param from_block: int or None
        :param to_block: int or None
        :return:
        """
        from contracts_lib_py.event_listener import EventListener
        return EventListener(
            self.CONTRACT_NAME,
            event_name,
            args,
            filters=event_filter,
            from_block=from_block,
            to_block=to_block
        ).listen_once(
            callback,
            timeout_callback=timeout_callback,
            timeout=timeout,
            blocking=wait
        )

    def send_transaction(self, fn_name, fn_args, transact=None):
        """Calls a smart contract function using either `personal_sendTransaction` (if
        passphrase is available) or `ether_sendTransaction`.

        :param fn_name: str the smart contract function name
        :param fn_args: tuple arguments to pass to function above
        :param transact: dict arguments for the transaction such as from, gas, etc.
        :return:
        """
        contract_fn = getattr(self.contract.functions, fn_name)(*fn_args)
        contract_function = CustomContractFunction(
            contract_fn
        )
        if transact is not None and 'chainId' not in transact:
            transact['chainId'] = int(Web3Provider.get_web3().net.version)
        return contract_function.transact(transact)

    def get_event_argument_names(self, event_name):
        event = getattr(self._contract.events, event_name, None)
        if event:
            return event().argument_names

    @property
    def function_names(self):
        return list(self._contract.function_names)

    def __str__(self):
        return f'{self.name} at {self.address}'



import logging
import time
from datetime import datetime
from threading import Thread

from contracts_lib_py.contract_handler import ContractHandler
from contracts_lib_py.event_filter import EventFilter

logger = logging.getLogger(__name__)


class EventListener(object):
    """Class representing an event listener."""

    def __init__(self, contract_name, event_name, args=None, from_block=None, to_block=None,
                 filters=None):
        contract = ContractHandler.get(contract_name)
        self.event_name = event_name
        self.event = getattr(contract.events, event_name)
        self.filters = filters if filters else {}
        self.from_block = from_block if from_block is not None else 'latest'
        self.to_block = to_block if to_block is not None else 'latest'
        self.event_filter = self.make_event_filter()
        self.event_filter.poll_interval = 0.5
        self.timeout = 600  # seconds
        self.args = args

    def make_event_filter(self):
        """Create a new event filter."""
        event_filter = EventFilter(
            self.event_name,
            self.event,
            self.filters,
            from_block=self.from_block,
            to_block=self.to_block
        )
        event_filter.set_poll_interval(0.5)
        return event_filter

    def listen_once(self, callback, timeout=None, timeout_callback=None, start_time=None,
                    blocking=False):
        """

        :param callback: a callback function that takes one argument the event dict
        :param timeout: float timeout in seconds
        :param timeout_callback: a callback function when timeout expires
        :param start_time: float start time in seconds, defaults to current time and is used
            for calculating timeout
        :param blocking: bool blocks this call until the event is detected
        :return: event if blocking is True and an event is received, otherwise returns None
        """
        if blocking:
            assert timeout is not None, '`timeout` argument is required when `blocking` is True.'

        events = []
        original_callback = callback

        def _callback(event, *args):
            events.append(event)
            if original_callback:
                original_callback(event, *args)

        if blocking:
            callback = _callback

        # TODO Review where to close this threads.
        Thread(
            target=self.watch_one_event,
            args=(self.event_filter,
                  callback,
                  timeout_callback,
                  timeout if timeout is not None else self.timeout,
                  self.args,
                  start_time),
            daemon=True,
        ).start()
        if blocking:
            while not events:
                time.sleep(0.2)

            return events[0]

        return None

    @staticmethod
    def watch_one_event(event_filter, callback, timeout_callback, timeout, args,
                        start_time=None):
        """
        Start to watch one event.

        :param event_filter:
        :param callback:
        :param timeout_callback:
        :param timeout:
        :param args:
        :param start_time:
        :return:
        """
        if timeout and not start_time:
            start_time = int(datetime.now().timestamp())

        if not args:
            args = []

        while True:
            try:
                events = event_filter.get_all_entries()
                if events:
                    callback(events[0], *args)
                    return

            except (ValueError, Exception) as err:
                # ignore error, but log it
                logger.debug(f'Got error grabbing keeper events: {str(err)}')

            time.sleep(0.5)
            if timeout:
                elapsed = int(datetime.now().timestamp()) - start_time
                if elapsed > timeout:
                    if timeout_callback is not None:
                        timeout_callback(*args)
                    elif callback is not None:
                        callback(None, *args)
                    break



from contracts_lib_py.conditions.condition_base import ConditionBase


class SignCondition(ConditionBase):
    """Class representing the SignCondition contract."""
    CONTRACT_NAME = 'SignCondition'

    def fulfill(self, agreement_id, message, account_address, signature, from_account):
        """
        Fulfill the sign conditon.

        :param agreement_id: id of the agreement, hex str
        :param message:
        :param account_address: ethereum account address, hex str
        :param signature: signed agreement hash, hex str
        :param from_account: Account doing the transaction
        :return:
        """
        return self._fulfill(
            agreement_id,
            message,
            account_address,
            signature,
            transact={'from': from_account.address,
                      'passphrase': from_account.password,
                      'keyfile': from_account.key_file}
        )

    def hash_values(self, message, account_address):
        """
        Hash the values of the sign condition.

        :param message:
        :param account_address:
        :return:
        """
        return self._hash_values(message, account_address)



import logging
import os

from contracts_lib_py import Keeper
from contracts_lib_py.contract_handler import ContractHandler
from contracts_lib_py.exceptions import ContractsNotFound

logger = logging.getLogger(__name__)


class Diagnostics:
    TEST_CONTRACT_NAME = 'AgreementStoreManager'

    @staticmethod
    def verify_contracts():
        """
        Verify that the contracts are deployed correctly in the network.

        :raise Exception: raise exception if the contracts are not deployed correctly.
        """
        artifacts_path = Keeper.get_instance().artifacts_path
        logger.info(f'Keeper contract artifacts (JSON abi files) at: {artifacts_path}')

        if os.environ.get('KEEPER_NETWORK_NAME'):
            logger.warning(f'The `KEEPER_NETWORK_NAME` env var is set to '
                           f'{os.environ.get("KEEPER_NETWORK_NAME")}. '
                           f'This enables the user to override the method of how the network name '
                           f'is inferred from network id.')

        # try to find contract with this network name
        contract_name = Diagnostics.TEST_CONTRACT_NAME
        network_id = Keeper.get_network_id()
        network_name = Keeper.get_network_name(network_id)
        logger.info(f'Using keeper contracts from network {network_name}, '
                    f'network id is {network_id}')
        logger.info(f'Looking for keeper contracts ending with ".{network_name}.json", '
                    f'e.g. "{contract_name}.{network_name}.json".')
        existing_contract_names = os.listdir(artifacts_path)
        try:
            ContractHandler.get(contract_name)
        except Exception as e:
            logger.error(e)
            logger.error(f'Cannot find the keeper contracts. \n'
                         f'Current network id is {network_id} and network name is {network_name}.'
                         f'Expected to find contracts ending with ".{network_name}.json",'
                         f' e.g. "{contract_name}.{network_name}.json"')
            raise ContractsNotFound(
                f'Keeper contracts for keeper network {network_name} were not found '
                f'in {artifacts_path}. \n'
                f'Found the following contracts: \n\t{existing_contract_names}'
            )

        keeper = Keeper.get_instance()
        contracts = [keeper.token, keeper.did_registry,
                     keeper.agreement_manager, keeper.template_manager, keeper.condition_manager,
                     keeper.access_condition, keeper.sign_condition,
                     keeper.lock_payment_condition, keeper.access_template,
                     keeper.escrow_payment_condition, keeper.hash_lock_condition
                     ]
        addresses = '\n'.join([f'\t{c.name}: {c.address}' for c in contracts])
        logging.info('Finished loading keeper contracts:\n'
                     '%s', addresses)



import logging

from contracts_lib_py.templates.template_base import TemplateBase

logger = logging.getLogger('didSalesTemplate')


class DIDSalesTemplate(TemplateBase):
    """Class representing the DIDSalesTemplate contract."""
    CONTRACT_NAME = 'DIDSalesTemplate'



class ContractsNotFound(Exception):
    """Raised when is not possible to find the keeper contracts abi."""


class DIDNotFound(Exception):
    """Raised when a requested DID or a DID in the chain cannot be found."""


class InvalidTransaction(Exception):
    """Raised when an on-chain transaction fail."""



from contracts_lib_py.conditions.condition_base import ConditionBase
from contracts_lib_py.event_filter import EventFilter


class AccessCondition(ConditionBase):
    """Class representing the AccessCondition contract."""
    CONTRACT_NAME = 'AccessCondition'

    def fulfill(self, agreement_id, document_id, grantee_address, account):
        """
        Fulfill the access condition.

        :param agreement_id: id of the agreement, hex str
        :param document_id: refers to the DID in which secret store will issue the decryption
        keys, DID
        :param grantee_address: is the address of the granted user, str
        :param account: Account instance
        :return: true if the condition was successfully fulfilled, bool
        """
        return self._fulfill(
            agreement_id,
            document_id,
            grantee_address,
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )

    def hash_values(self, document_id, grantee_address):
        """
        Hast the values of the document_id with the grantee address.

        :param document_id: refers to the DID in which secret store will issue the decryption
        keys, DID
        :param grantee_address: is the address of the granted user, str
        :return: hex str
        """
        return self._hash_values(document_id, grantee_address)

    def check_permissions(self, document_id, grantee_address):
        """
        Check that the grantee_address has permissions to decrypt the document stored with this
        document_id.

        :param document_id: refers to the DID in which secret store will issue the decryption
        keys, DID
        :param grantee_address: is the address of the granted user, str
        :return: true if the access was granted, bool
        """
        return self.contract.caller.checkPermissions(grantee_address, document_id)

    def get_purchased_assets_by_address(self, address, from_block=0, to_block='latest'):
        """
        Get the list of the assets dids consumed for an address.

        :param address: is the address of the granted user, hex-str
        :param from_block: block to start to listen
        :param to_block: block to stop to listen
        :return: list of dids
        """
        block_filter = EventFilter(
            ConditionBase.FULFILLED_EVENT,
            getattr(self.events, ConditionBase.FULFILLED_EVENT),
            from_block=from_block,
            to_block=to_block,
            argument_filters={'_grantee': address}
        )
        log_items = block_filter.get_all_entries(max_tries=5)
        did_list = []
        for log_i in log_items:
            did_list.append(log_i.args['_documentId'])

        return did_list



import logging

logger = logging.getLogger(__name__)


class Wallet:
    """
    The wallet is responsible for signing transactions and messages by using an account's
    private key.

    The private key is always red from the encrypted keyfile and is never saved in memory beyond
    the life span of the signing function.

    The use of this wallet allows Nevermined tools to send rawTransactions which keeps the user
    key and password safe and they are never sent outside. Another advantage of this is that
    we can interact directly with remote keeper nodes without having to run a local parity
    node since we only send the raw transaction hash so the user info is safe.

    Note that currently a local node is still necessary for completing the consume flow because
    the secretstore needs a parity node to encrypt/decrypt the asset urls. We could eliminate
    this dependency on local parity node by using Gateway's publish endpoint for encryption and
    sending url index to Gateway's consume endpoint instead of decrypting locally first.
    The secretstore also uses the parity node to sign the document key id which is necessary
    for communicating with the secret store nodes.

    """
    _last_tx_count = dict()
    MIN_GAS_PRICE = 1000000000

    def __init__(self, web3, key_file, password, address=None):
        self._web3 = web3
        self._key_file = key_file
        self._password = password
        self._address = address

    def __read_key(self):
        with open(self._key_file) as _file:
            encrypted_key = _file.read()
            private_key = self._web3.eth.account.decrypt(encrypted_key, self._password)
            return private_key

    def validate(self):
        key = self.__read_key()
        account = self._web3.eth.account.from_key(key)
        return account.address == self._address

    @staticmethod
    def _get_nonce(web3, address):
        # We cannot rely on `web3.eth.getTransactionCount` because when sending multiple
        # transactions in a row without wait in between the network may not get the chance to
        # update the transaction count for the account address in time.
        # So we have to manage this internally per account address.
        if address not in Wallet._last_tx_count:
            Wallet._last_tx_count[address] = web3.eth.getTransactionCount(address)
        else:
            Wallet._last_tx_count[address] += 1

        return Wallet._last_tx_count[address]

    def sign_tx(self, tx):
        private_key = self.__read_key()
        account = self._web3.eth.account.from_key(private_key)
        # nonce = Wallet._get_nonce(self._web3, account.address)
        # logger.debug(f'`Wallet` signing tx: sender address: {account.address} nonce: {nonce}, '
        #              f'gasprice: {self._web3.eth.gasPrice}')
        signed_tx = 0
        try:
            priority_fee = self._web3.eth.max_priority_fee
            gas_price = self._web3.eth.gas_price
            tx['type'] = 2
            tx['nonce'] = self._web3.eth.getTransactionCount(account.address)
            tx['maxPriorityFeePerGas'] = priority_fee
            tx['maxFeePerGas'] = gas_price
            signed_tx = self._web3.eth.account.sign_transaction(tx, private_key)
        except:
            gas_price = self._web3.eth.gas_price
            tx['nonce'] = self._web3.eth.getTransactionCount(account.address)
            tx['gasPrice'] = gas_price
            signed_tx = self._web3.eth.account.sign_transaction(tx, private_key)
        logger.debug(f'`Wallet` signed tx is {signed_tx}')
        return signed_tx.rawTransaction

    def sign(self, msg_hash):
        private_key = self.__read_key()
        account = self._web3.eth.account.from_key(private_key)
        return account.signHash(msg_hash)



from web3 import Web3

from contracts_lib_py.contract_base import ContractBase


class Token(ContractBase):
    """Class representing the Token contract."""
    CONTRACT_NAME = 'NeverminedToken'

    def get_token_balance(self, account_address):
        """
        Retrieve the amount of tokens of an account address.

        :param account_address: Account address, str
        :return: int
        """
        return self.contract.caller.balanceOf(account_address)

    def get_allowance(self, owner_address, spender_address):
        """

        :param owner_address: Address, str
        :param spender_address: Address, str
        :return:
        """
        return self.contract.caller.allowance(owner_address, spender_address)

    def token_approve(self, spender_address, price, from_account):
        """
        Approve the passed address to spend the specified amount of tokens.

        :param spender_address: Account address, str
        :param price: Asset price, int
        :param from_account: Account address, str
        :return: bool
        """
        if not Web3.isChecksumAddress(spender_address):
            spender_address = Web3.toChecksumAddress(spender_address)

        tx_hash = self.send_transaction(
            'approve',
            (spender_address,
             price),
            transact={'from': from_account.address,
                      'passphrase': from_account.password,
                      'keyfile': from_account.key_file}
        )
        return self.is_tx_successful(tx_hash)

    def transfer(self, receiver_address, amount, from_account):
        """
        Transfer tokens from one account to the receiver address.

        :param receiver_address: Address of the transfer receiver, str
        :param amount: Amount of tokens, int
        :param from_account: Sender account, Account
        :return: bool
        """
        tx_hash = self.send_transaction(
            'transfer',
            (receiver_address,
             amount),
            transact={'from': from_account.address,
                      'passphrase': from_account.password,
                      'keyfile': from_account.key_file}
        )
        return self.is_tx_successful(tx_hash)

    def total_supply(self):
        """
        Return the total supply

        :return: int
        """
        return self.contract.caller.totalSupply()

    def increase_allowance(self, spender_address, added_value, owner_account):
        """

        :param spender_address:
        :param added_value:
        :param owner_account:
        :return: bool
        """
        tx_hash = self.send_transaction(
            'increaseAllowance',
            (spender_address,
             added_value),
            transact={'from': owner_account.address,
                      'passphrase': owner_account.password,
                      'keyfile': owner_account.key_file}
        )
        return self.is_tx_successful(tx_hash)

    def decrease_allowance(self, spender_address, subtracted_value, owner_account):
        """

        :param spender_address:
        :param subtracted_value:
        :param owner_account:
        :return: bool
        """
        tx_hash = self.send_transaction(
            'decreaseAllowance',
            (spender_address,
             subtracted_value),
            transact={'from': owner_account.address,
                      'passphrase': owner_account.password,
                      'keyfile': owner_account.key_file}
        )
        return self.is_tx_successful(tx_hash)



from contracts_lib_py.conditions.condition_base import ConditionBase


class ComputeExecutionCondition(ConditionBase):
    """Class representing the ComputeExecutionCondition contract."""
    CONTRACT_NAME = 'ComputeExecutionCondition'

    def fulfill(self, agreement_id, did, computer_consumer_address, account):
        """
        Fulfill the compute execution condition.

        :param agreement_id: id of the agreement, hex str
        :param did: the id of an asset on-chain, hex str
        :param computer_consumer_address: is the address of the consumer user, str
        :param account: Account instance
        :return: true if the condition was successfully fulfilled, bool
        """
        return self._fulfill(
            agreement_id,
            did,
            computer_consumer_address,
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )

    def hash_values(self, did, computer_consumer_address):
        """
        Hash the values of the compute execution condition.

        :param did: the id of an asset on-chain, hex str
        :param computer_consumer_address: is the address of the consumer user, str
        :return: hex str
        """
        return self._hash_values(did, computer_consumer_address)

    def was_compute_triggered(self, did, computer_consumer_address):
        """
        Checks whether the compute is triggered or not.

        :param did: the id of an asset on-chain, hex str
        :param computer_consumer_address: is the address of the consumer user, str
        :return: bool
        """
        return self.contract.caller.wasComputeTriggered(did, computer_consumer_address)



from contracts_lib_py.conditions.condition_base import ConditionBase


class NFTHolderCondition(ConditionBase):
    """Class representing the NFTHolderCondition contract."""
    CONTRACT_NAME = 'NFTHolderCondition'

    def fulfill(self, agreement_id, did, holder_address, amount, account):
        """
        Fulfill the NFT Holder condition.

        :param agreement_id: id of the agreement, hex str
        :param did: refers to the DID in which secret store will issue the decryption
        keys, DID
        :param holder_address: is the address of the user holding the NFT, str
        :param amount: number of NFTs to hold, str
        :param account: Account instance
        :return: true if the condition was successfully fulfilled, bool
        """
        return self._fulfill(
            agreement_id,
            did,
            holder_address,
            amount,
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )

    def hash_values(self, did, holder_address, amount):
        """
        Hast the values of the document_id with the grantee address.

        :param did: refers to the DID in which secret store will issue the decryption
        keys, DID
        :param holder_address: is the address of the user holding the NFT, str
        :param amount: number of NFTs to hold, str
        :return: hex str
        """
        return self._hash_values(did, holder_address, amount)





from collections import namedtuple

from contracts_lib_py import ContractBase

ConditionValues = namedtuple(
    'ConditionValues',
    ('type_ref', 'state', 'time_lock', 'time_out', 'block_number')
)


class ConditionStoreManager(ContractBase):
    """Class representing the ConditionStoreManager contract."""
    CONTRACT_NAME = 'ConditionStoreManager'

    def get_condition(self, condition_id):
        """Retrieve the condition for a condition_id.

        :param condition_id: id of the condition, str
        :return:
        """
        condition = self.contract.caller.getCondition(condition_id)
        if condition and len(condition) == 5:
            return ConditionValues(*condition)

        return None

    def get_condition_state(self, condition_id):
        """Retrieve the condition state.

        :param condition_id: id of the condition, str
        :return: State of the condition
        """
        return self.contract.caller.getConditionState(condition_id)

    def get_num_condition(self):
        """
        Return the size of the Conditions list.

        :return: the length of the conditions list, int
        """
        return self.contract.caller.getConditionListSize()



import logging
import time

from contracts_lib_py.web3_provider import Web3Provider

logger = logging.getLogger(__name__)


class EventFilter:
    def __init__(self, event_name, event, argument_filters, from_block, to_block,
                 poll_interval=None):
        self.event_name = event_name
        self.event = event
        self.argument_filters = argument_filters
        self.block_range = (from_block, to_block)
        self._filter = None
        self._poll_interval = poll_interval if poll_interval else 0.5
        self._create_filter()

    @property
    def filter_id(self):
        return self._filter.filter_id if self._filter else None

    def uninstall(self):
        Web3Provider.get_web3().eth.uninstallFilter(self._filter.filter_id)

    def set_poll_interval(self, interval):
        self._poll_interval = interval

    def recreate_filter(self):
        self._create_filter()

    def _create_filter(self):
        chain_id = Web3Provider.get_web3().net.version
        from_block = 0

        # temporary workaround to work with mumbai
        # Infura has a 1000 block range limit in their api
        if chain_id == '80001':
            latest_block = Web3Provider.get_web3().eth.get_block_number()
            from_block = latest_block - 990


        self._filter = {
            'fromBlock': from_block,
            'toBlock': self.block_range[1],
            'argument_filters': self.argument_filters
        }

    def get_new_entries(self, max_tries=1):
        # This api is not provided by polygon
        try:
            return self._get_entries(max_tries=max_tries)
        except ValueError as e:
            # method does not exist or is not available
            if e.args[0]['code'] == -32601:
                return []
            else:
                raise

    def get_all_entries(self, max_tries=1):
         # This api is not provided by polygon
        try:
            return self._get_entries(max_tries=max_tries)
        except ValueError as e:
            # method does not exist or is not available
            if e.args[0]['code'] == -32601:
                return []
            else:
                raise

    def _get_entries(self, max_tries=1):
        i = 0
        while i < max_tries:
            try:
                logs = self.event.getLogs(**self._filter)
                if logs:
                    logger.debug(
                        f'found event logs: event-name={self.event_name}, '
                        f'range={self.block_range}, '
                        f'logs={logs}')
                    return logs
            except ValueError as e:
                if 'Filter not found' in str(e):
                    logger.debug(f'recreating filter (Filter not found): event={self.event_name}, '
                                 f'arg-filter={self.argument_filters}, from/to={self.block_range}')
                    time.sleep(1)
                    self._create_filter()
                else:
                    raise

            i += 1
            if max_tries > 1 and i < max_tries:
                time.sleep(0.5)

        return []



from contracts_lib_py.conditions.condition_base import ConditionBase

class NFTLockCondition(ConditionBase):
    """Class representing the NFTLockCondition contract."""
    CONTRACT_NAME = 'NFTLockCondition'

    def fulfill(self, agreement_id, document_id, reward_address, amount, account):
        """
        Fulfill the nft lock condition.

        :param agreement_id: id of the agreement, hex str
        :param document_id: refers to the DID in which secret store will issue the decryption
        keys, DID
        :param reward_address: is the contract address where the reward is locked, str
        :param amount: is the amount of tokens to be transfered, int
        :param account: Account instance
        :return: true if the condition was successfully fulfilled, bool
        """
        return self._fulfill(
            agreement_id,
            document_id,
            reward_address,
            amount,
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )

    def fulfill_marked(self, agreement_id, document_id, reward_address, amount, receiver, token_address, account):
        return self._fulfill(
            agreement_id,
            document_id,
            reward_address,
            amount,
            receiver,
            token_address,
            method='fulfillMarked',
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )

    def hash_values(self, document_id, reward_address, amount):
        """
        Hash the values of the document_id with the reward_address and amount.

        :param document_id: refers to the DID in which secret store will issue the decryption
        keys, DID
        :param reward_address: is the final address to receive the NFTs, str
        :param amount: is the amount of locked tokens, int
        :return: hex str
        """
        return self._hash_values(document_id, reward_address, amount)

    def hash_values_marked(self, document_id, reward_address, amount, receiver, token_address):
        return self._hash_values_marked(document_id, reward_address, amount, receiver, token_address)



from .access import AccessCondition
from .escrow_reward import EscrowPaymentCondition
from .hash_lock import HashLockCondition
from .lock_reward import LockPaymentCondition
from .sign import SignCondition



"""Test NFTLockCondition contract."""
from contracts_lib_py.conditions.nft_lock_condition import NFTLockCondition


def test_nft_lock_condition_contract():
    nft_lock_condition = NFTLockCondition('NFTLockCondition')
    assert nft_lock_condition
    assert isinstance(nft_lock_condition, NFTLockCondition), \
        f'{nft_lock_condition} is not instance of NFTLockCondition'


import logging

from web3._utils import empty
from web3.contract import prepare_transaction

from contracts_lib_py.wallet import Wallet


class CustomContractFunction:

    def __init__(self, contract_function):
        self._contract_function = contract_function

    def transact(self, transaction=None):
        """
        Customize calling smart contract transaction functions to use `sendTransaction`
        instead of `eth_sendTransaction` and to estimate gas limit. This function
        is largely copied from web3 ContractFunction with important addition.

        Note: will fallback to `eth_sendTransaction` if `passphrase` is not provided in the
        `transaction` dict.

        :param transaction: dict which has the required transaction arguments per
            `sendTransaction` requirements.
        :return: hex str transaction hash
        """
        if transaction is None:
            transact_transaction = {}
        else:
            transact_transaction = dict(**transaction)

        if 'data' in transact_transaction:
            raise ValueError("Cannot set data in transact transaction")

        cf = self._contract_function
        if cf.address is not None:
            transact_transaction.setdefault('to', cf.address)
        if cf.web3.eth.defaultAccount is not empty:
            transact_transaction.setdefault('from', cf.web3.eth.defaultAccount)

        if 'to' not in transact_transaction:
            if isinstance(self, type):
                raise ValueError(
                    "When using `Contract.transact` from a contract factory you "
                    "must provide a `to` address with the transaction"
                )
            else:
                raise ValueError(
                    "Please ensure that this contract instance has an address."
                )

        if 'gas' not in transact_transaction:
            tx = transaction.copy()
            if 'passphrase' in tx:
                tx.pop('passphrase')
            if 'keyfile' in tx:
                tx.pop('keyfile')
            gas = cf.estimateGas(tx)
            transact_transaction['gas'] = gas

        return transact_with_contract_function(
            cf.address,
            cf.web3,
            cf.function_identifier,
            transact_transaction,
            cf.contract_abi,
            cf.abi,
            *cf.args,
            **cf.kwargs
        )


def transact_with_contract_function(
        address,
        web3,
        function_name=None,
        transaction=None,
        contract_abi=None,
        fn_abi=None,
        *args,
        **kwargs):
    """
    Helper function for interacting with a contract function by sending a
    transaction. This is copied from web3 `transact_with_contract_function`
    so we can use `personal_sendTransaction` when possible.
    """
    transact_transaction = prepare_transaction(
        address,
        web3,
        fn_identifier=function_name,
        contract_abi=contract_abi,
        transaction=transaction,
        fn_abi=fn_abi,
        fn_args=args,
        fn_kwargs=kwargs,
    )

    passphrase = None
    key_file = None
    if transaction and 'passphrase' in transaction:
        passphrase = transaction['passphrase']
        transact_transaction.pop('passphrase')
        if 'keyfile' in transaction:
            key_file = transaction['keyfile']
            transact_transaction.pop('keyfile')

    # Restrict transactions to raw transactions for now (security first)
    # if not (passphrase and key_file):
    #     raise AssertionError(
    #         'password and key file are required for signing transactions locally.'
    #     )

    if passphrase and key_file:
        raw_tx = Wallet(web3, key_file, passphrase).sign_tx(transact_transaction)
        logging.debug(f'sending raw tx: function: {function_name}, tx hash: {raw_tx.hex()}')
        txn_hash = web3.eth.sendRawTransaction(raw_tx)
    elif passphrase:
        txn_hash = web3.sendTransaction(transact_transaction, passphrase)
    else:
        txn_hash = web3.eth.sendTransaction(transact_transaction)

    return txn_hash



import logging
import os

logger = logging.getLogger('account')


class Account:
    """Class representing an account."""

    def __init__(self, address, password=None, key_file=None):
        """
        Hold account address, password and path to keyfile.

        :param address: The address of this account
        :param password: account's password. This is necessary for decrypting the private key
            to be able to sign transactions locally
        :param key_file: str path to the encrypted private key file
        """
        self.address = address
        self.password = password
        self._key_file = key_file

    @property
    def key_file(self):
        if self._key_file:
            return os.path.expandvars(os.path.expanduser(self._key_file))
        return None



import json
import logging
import os

from web3 import Web3

from contracts_lib_py import Keeper
from contracts_lib_py.web3_provider import Web3Provider

logger = logging.getLogger(__name__)


class ContractHandler(object):
    """
    Manages loading contracts and also keeps a cache of loaded contracts.

    Retrieval of deployed keeper contracts must use this `ContractHandler`.
    Example:
        contract = ContractHandler.get('ServiceExecutionAgreement')
    """
    _contracts = dict()
    artifacts_path = None

    @staticmethod
    def get(name):
        """
        Return the Contract instance for a given name.

        :param name: Contract name, str
        :return: Contract instance
        """
        return (ContractHandler._contracts.get(name) or ContractHandler._load(name))[0]

    @staticmethod
    def get_by_address(address, abi, name):
        """
        Return the Contract instance for a given name.

        :param name: Contract name, str
        :return: Contract instance
        """
        return (ContractHandler._contracts.get(name) or ContractHandler._load_from_address(address, abi, name))[0]

    @staticmethod
    def get_contract_version(name):
        """
        Return the version of the contract in use.

        :param name: name of the contract
        :return: str version
        """
        return ContractHandler._contracts.get(name)[1]

    @staticmethod
    def set(name, contract):
        """
        Set a Contract instance for a contract name.

        :param name: Contract name, str
        :param contract: Contract instance
        """
        ContractHandler._contracts[name] = contract

    @staticmethod
    def has(name):
        """
        Check if a contract is the ContractHandler contracts.

        :param name: Contract name, str
        :return: True if the contract is there, bool
        """
        return name in ContractHandler._contracts

    @staticmethod
    def _load(contract_name):
        """Retrieve the contract instance for `contract_name` that represent the smart
        contract in the keeper network.

        :param contract_name: str name of the solidity keeper contract without the network name.
        :return: web3.eth.Contract instance
        """
        assert ContractHandler.artifacts_path is not None, 'artifacts_path should be already set.'
        contract_definition = ContractHandler.get_contract_dict_by_name(
            contract_name, ContractHandler.artifacts_path)
        address = Web3.toChecksumAddress(contract_definition['address'])
        abi = contract_definition['abi']
        contract = Web3Provider.get_web3().eth.contract(address=address, abi=abi)
        ContractHandler._contracts[contract_name] = (contract, contract_definition['version'])
        return ContractHandler._contracts[contract_name]

    @staticmethod
    def _load_from_address(address, abi, contract_name):
        """Retrieve the contract instance for `contract_name` that represent the smart
        contract in the keeper network.

        :param contract_name: str name of the solidity keeper contract without the network name.
        :return: web3.eth.Contract instance
        """
        address = Web3.toChecksumAddress(address)
        contract = Web3Provider.get_web3().eth.contract(address=address, abi=abi)
        ContractHandler._contracts[contract_name] = (contract, 'external')
        return ContractHandler._contracts[contract_name]

    @staticmethod
    def _get_contract_file_path(_base_path, _contract_name, _network_name, artifacts_path):
        contract_file_name = '{}.{}.json'.format(_contract_name, _network_name)
        for name in os.listdir(_base_path):
            if name.lower() == contract_file_name.lower():
                contract_file_name = name
                return os.path.join(artifacts_path, contract_file_name)
        return None

    @staticmethod
    def get_contract_dict_by_name(contract_name, artifacts_path):
        """
        Retrieve the Contract instance for a given contract name.

        :param contract_name: str
        :param artifacts_path: str the path to keeper contracts artifacts (`abi` .json files)
        :return: the smart contract's definition from the json abi file, dict
        """

        network_name = Keeper.get_network_name(Keeper.get_network_id()).lower()

        # file_name = '{}.{}.json'.format(contract_name, network_name)
        # path = os.path.join(keeper.artifacts_path, file_name)
        path = ContractHandler._get_contract_file_path(
            artifacts_path, contract_name, network_name, artifacts_path)
        if not (path and os.path.exists(path)):
            path = ContractHandler._get_contract_file_path(
                artifacts_path, contract_name, network_name.lower(), artifacts_path)

        if not (path and os.path.exists(path)):
            path = ContractHandler._get_contract_file_path(
                artifacts_path, contract_name, Keeper.DEFAULT_NETWORK_NAME, artifacts_path)

        if not (path and os.path.exists(path)):
            raise FileNotFoundError(
                f'Keeper contract {contract_name} file '
                f'not found in {artifacts_path} '
                f'using network name {network_name}'
            )

        with open(path) as f:
            contract_dict = json.loads(f.read())
            return contract_dict



import logging

from contracts_lib_py.templates.template_base import TemplateBase

logger = logging.getLogger('AccessProofTemplate')


class NFTAccessProofTemplate(TemplateBase):
    """Class representing the AccessProofTemplate contract."""
    CONTRACT_NAME = 'NFTAccessProofTemplate'



from web3 import Web3

from contracts_lib_py.web3.http_provider import CustomHTTPProvider


class Web3Provider(object):
    """Provides the Web3 instance."""
    _web3 = None

    @staticmethod
    def get_web3(keeper_url=None, provider=None):
        """Return the web3 instance to interact with the ethereum client."""
        if Web3Provider._web3 is None:
            if not provider:
                assert keeper_url, 'keeper_url or a provider instance is required.'
                provider = CustomHTTPProvider(keeper_url)

            Web3Provider._web3 = Web3(provider)
            # Reset attributes to avoid lint issue about no attribute
            Web3Provider._web3.EthereumTesterProvider = getattr(Web3Provider._web3, 'EthereumTesterProvider')
            Web3Provider._web3.HTTPProvider = getattr(Web3Provider._web3, 'HTTPProvider')
            Web3Provider._web3.IPCProvider = getattr(Web3Provider._web3, 'IPCProvider')
            Web3Provider._web3.Iban = getattr(Web3Provider._web3, 'Iban')
            Web3Provider._web3.RequestManager = getattr(Web3Provider._web3, 'RequestManager')
            Web3Provider._web3.WebsocketProvider = getattr(Web3Provider._web3, 'WebsocketProvider')
            Web3Provider._web3.codec = getattr(Web3Provider._web3, 'codec')
            Web3Provider._web3.eth = getattr(Web3Provider._web3, 'eth')
            Web3Provider._web3.geth = getattr(Web3Provider._web3, 'geth')
            Web3Provider._web3.manager = getattr(Web3Provider._web3, 'manager')
            Web3Provider._web3.net = getattr(Web3Provider._web3, 'net')
            Web3Provider._web3.parity = getattr(Web3Provider._web3, 'parity')
            Web3Provider._web3.provider = getattr(Web3Provider._web3, 'provider')
            Web3Provider._web3.testing = getattr(Web3Provider._web3, 'testing')
            Web3Provider._web3.version = getattr(Web3Provider._web3, 'version')

        return Web3Provider._web3

    @staticmethod
    def set_web3(web3):
        Web3Provider._web3 = web3



"""Test Token Contract."""
import pytest
from web3 import Web3

from contracts_lib_py.conditions import LockPaymentCondition
from contracts_lib_py.token import Token
from contracts_lib_py.keeper import Keeper
from tests.resources.helper_functions import get_consumer_account, get_publisher_account

consumer_account = get_consumer_account()
publisher_account = get_publisher_account()


def test_token_contract():
    token = Token.get_instance()
    assert token
    assert isinstance(token, Token)


def test_get_balance():
    token = Token.get_instance()
    assert isinstance(token.get_token_balance(consumer_account.address), int)


def test_get_balance_invalid_address():
    token = Token.get_instance()
    with pytest.raises(Exception):
        token.get_token_balance('not valid')


def test_token_approve():
    token = Token.get_instance()
    assert token.token_approve(consumer_account.address, 100, publisher_account)


def test_token_approve_invalid_address():
    token = Token.get_instance()
    with pytest.raises(Exception):
        token.token_approve('10923019', 100, publisher_account)


def test_token_approve_invalid_tokens():
    token = Token.get_instance()
    with pytest.raises(Exception):
        token.token_approve(consumer_account.address, -100, publisher_account)


def test_token_allowance():
    token = Token.get_instance()
    lock_reward_condition = LockPaymentCondition(LockPaymentCondition.CONTRACT_NAME)
    allowance = token.get_allowance(consumer_account.address, lock_reward_condition.address)
    if allowance > 0:
        token.decrease_allowance(lock_reward_condition.address, allowance, consumer_account)
        allowance = token.get_allowance(consumer_account.address, lock_reward_condition.address)
    assert allowance == 0

    assert token.token_approve(lock_reward_condition.address, 77, consumer_account) is True
    allowance = token.get_allowance(consumer_account.address, lock_reward_condition.address)
    assert allowance == 77

    assert token.token_approve(lock_reward_condition.address, 49, consumer_account) is True
    allowance = token.get_allowance(consumer_account.address, lock_reward_condition.address)
    assert allowance == 49

    token.decrease_allowance(lock_reward_condition.address, 5, consumer_account)
    allowance = token.get_allowance(consumer_account.address, lock_reward_condition.address)
    assert allowance == 44

    token.increase_allowance(lock_reward_condition.address, 10, consumer_account)
    allowance = token.get_allowance(consumer_account.address, lock_reward_condition.address)
    assert allowance == 54


def test_token_transfer():
    test_address = Web3.toChecksumAddress('068ed00cf0441e4829d9784fcbe7b9e26d4bd8d0')
    token = Token.get_instance()
    balance = token.get_token_balance(test_address)
    token.transfer(test_address, 3, consumer_account)
    new_balance = token.get_token_balance(test_address)
    assert new_balance == (balance + 3)



from contracts_lib_py.conditions.condition_base import ConditionBase


class ThresholdCondition(ConditionBase):
    """Class representing the ThresholdCondition contract."""
    CONTRACT_NAME = 'ThresholdCondition'

    def fulfill(self, agreement_id, input_conditions, threshold, from_account):
        """
        Fulfill the Threshold condition

        :param agreement_id: id of the agreement, hex str
        :param input_conditions: array of input conditions IDs
        :param threshold: the required number of fulfilled input conditions, int
        :param from_account: Account instance
        :return:
        """
        assert len(input_conditions) >= 2, 'The minimum number of conditions is 2.'
        assert threshold <= len(input_conditions), \
            'The required number of condition to be fulfilled(threshold) should ' \
            'be smaller than the number of conditions.'
        return self._fulfill(
            agreement_id,
            input_conditions,
            threshold,
            transact={'from': from_account.address,
                      'passphrase': from_account.password,
                      'keyfile': from_account.key_file}
        )

    def hash_values(self, input_conditions, threshold):
        """
        HashValues generates the hash of condition inputs with the following parameters

        :param input_conditions: array of input conditions IDs
        :param threshold: the required number of fulfilled input conditions, int
        :return: hex str
        """
        return self._hash_values(input_conditions, threshold)

    def can_fulfill(self, input_conditions, threshold):
        """
        Check if condition can be fulfilled

        :param input_conditions: array of input conditions IDs
        :param threshold: the required number of fulfilled input conditions, int
        :return:
        """
        return self.contract.caller.canFulfill(input_conditions, threshold)



import logging
from collections import namedtuple
from urllib.parse import urlparse

from eth_utils import add_0x_prefix
from web3 import Web3
from web3._utils.events import get_event_data

from contracts_lib_py.contract_base import ContractBase
from contracts_lib_py.event_filter import EventFilter
from contracts_lib_py.exceptions import DIDNotFound
from contracts_lib_py.nft_upgradeable import NFTUpgradeable
from contracts_lib_py.web3_provider import Web3Provider

logger = logging.getLogger(__name__)

DIDRegisterValues = namedtuple(
    'DIDRegisterValues',
    ('owner', 'last_checksum', 'url', 'last_updated_by', 'block_number_updated', 'providers',
     'nft_supply', 'mint_cap', 'royalties')
)


class DIDRegistry(ContractBase):
    """Class to register and update DID's."""
    DID_REGISTRY_EVENT_NAME = 'DIDAttributeRegistered'
    PROVENANCE_ATTRIBUTE_REGISTERED_EVENT_NAME = 'ProvenanceAttributeRegistered'
    PROVENANCE_USED_EVENT_NAME = 'Used'
    PROVENANCE_WAS_GENERATED_BY_EVENT_NAME = 'WasGeneratedBy'
    PROVENANCE_WAS_DERIVED_FROM_EVENT_NAME = 'WasDerivedFrom'
    PROVENANCE_WAS_ASSOCIATED_WITH_EVENT_NAME = 'WasAssociatedWith'
    PROVENANCE_ACTED_ON_BEHALF_EVENT_NAME = 'ActedOnBehalf'

    CONTRACT_NAME = 'DIDRegistry'

    def register_mintable_did(self, did_seed, checksum, url, cap, royalties, account, providers=None, activity_id=None,
                              attributes=None):
        """
        Register a mintable DID using the DIDRegistry smart contract.

        :param did_seed: Seed used to generate the final DID, It's a 32 byte or hexstring
        :param checksum: hex str hash of TODO
        :param url: URL of the resolved DID
        :param account: instance of Account to use to register/update the DID
        :param cap: refers to the mint cap
        :param royalties: refers to the royalties to reward to the DID creator in the secondary market
        :param providers: list of addresses of providers to be allowed to serve the asset and play
            a part in creating and fulfilling service agreements
        :param activity_id: id of the activity tracked in the provenance.
        :param attributes: additional provenance attributes
        :return: Receipt
        """
        return self.register(did_seed, checksum, url, account,
                             cap=cap, royalties=royalties, attributes=attributes,
                             providers=providers, activity_id=activity_id)

    def register(self, did_seed, checksum, url, account, providers=None, activity_id=None,
                 attributes=None, cap=None, royalties=None):
        """
        Register or update a DID on the block chain using the DIDRegistry smart contract.

        :param did_seed: Seed used to generate the final DID, It's a 32 byte or hexstring
        :param checksum: hex str hash of TODO
        :param url: URL of the resolved DID
        :param account: instance of Account to use to register/update the DID
        :param providers: list of addresses of providers to be allowed to serve the asset and play
            a part in creating and fulfilling service agreements
        :param activity_id: id of the activity tracked in the provenance.
        :param cap: refers to the mint cap
        :param royalties: refers to the royalties to reward to the DID creator in the secondary market
        :param attributes: additional provenance attributes
        :return: Receipt
        """
        if isinstance(did_seed, bytes):
            did_source_id = did_seed
        elif isinstance(did_seed, str):
            try:
                did_source_id = Web3.toBytes(hexstr=did_seed)
            except Exception as e:
                raise TypeError(
                    f'There is a problem with the did type, expecting a str with valid hex only, '
                    f'got {did_seed}. Make sure to use only the id part of the '
                    f'did without the prefix: {e}')
        else:
            raise TypeError(f'Unrecognized `did_seed` {did_seed} of type {type(did_seed)}. Expecting a 32 bytes '
                            f'in bytes type or hex str representation.')

        if not did_source_id:
            raise ValueError(f'{did_seed} must be a valid DID to register')

        if not urlparse(url):
            raise ValueError(f'Invalid URL {url} to register for DID {did_seed}')

        if checksum is None:
            checksum = Web3.toBytes(0)

        if not isinstance(checksum, bytes):
            raise ValueError(f'Invalid checksum value {checksum}, must be bytes or string')

        if account is None:
            raise ValueError('You must provide an account to use to register a DID')

        if cap is not None or royalties is not None:
            transaction = self._register_mintable_did(
                did_source_id, checksum, providers or [], url, account, cap, royalties, activity_id, attributes
            )
        else:
            transaction = self._register_did(
                did_source_id, checksum, providers or [], url, account, activity_id, attributes
            )
        receipt = self.get_tx_receipt(transaction)
        if receipt:
            return receipt.status == 1

        did = self.hash_did(did_source_id, account.address)

        _filters = dict()
        _filters['_did'] = Web3.toBytes(hexstr=did)
        _filters['_owner'] = Web3.toBytes(hexstr=account.address)
        event = self.subscribe_to_event(self.DID_REGISTRY_EVENT_NAME, 15, _filters, wait=True)
        return event is not None

    def hash_did(self, did_seed, address):
        """
        Calculates the final DID given a DID seed and the publisher address
        :param did_seed: the hash used to generate the did
        :param address: the address of the account creating the asset
        :return: the final DID
        """
        return add_0x_prefix(self.contract.caller.hashDID(did_seed, address).hex())

    def are_royalties_valid(self, did, amounts, receivers):
        """
        Validates if asset royalties are valid
        :param did:  Asset did, did
        :param amounts: amounts to distribute
        :param receivers: receivers of the amounts
        :return: boolean
        """
        return self.contract.caller.areRoyaltiesValid(did, amounts, receivers)

    def get_block_number_updated(self, did):
        """Return the block number the last did was updated on the block chain."""
        return self.contract.caller.getBlockNumberUpdated(did)

    def get_did_owner(self, did):
        """
        Return the owner of the did.

        :param did: Asset did, did
        :return: ethereum address, hex str
        """
        return self.contract.caller.getDIDOwner(did)

    def add_provider(self, did, provider_address, account):
        """
        Add the provider of the did.

        :param did: the id of an asset on-chain, hex str
        :param provider_address: ethereum account address of the provider, hex str
        :param account: Account executing the action
        :return:
        """
        tx_hash = self.send_transaction(
            'addDIDProvider',
            (did,
             provider_address),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )
        return self.is_tx_successful(tx_hash)

    def remove_provider(self, did, provider_address, account):
        """
        Remove a provider

        :param did: the id of an asset on-chain, hex str
        :param provider_address: ethereum account address of the provider, hex str
        :param account: Account executing the action
        :return:
        """
        tx_hash = self.send_transaction(
            'removeDIDProvider',
            (did,
             provider_address),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )
        return self.is_tx_successful(tx_hash)

    def transfer_did_ownership(self, did, new_owner_address, account):
        """
        Transfer did ownership to an address.

        :param did: the id of an asset on-chain, hex str
        :param new_owner_address: ethereum account address, hex str
        :param account: Account executing the action
        :return: bool
        """
        tx_hash = self.send_transaction(
            'transferDIDOwnership',
            (did, new_owner_address),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )
        return self.is_tx_successful(tx_hash)

    def grant_permission(self, did, address_to_grant, account):
        """
        Grant access permission to an address.

        :param did: the id of an asset on-chain, hex str
        :param address_to_grant: ethereum account address, hex str
        :param account: Account executing the action
        :return: bool
        """
        tx_hash = self.send_transaction(
            'grantPermission',
            (did, address_to_grant),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )
        return self.is_tx_successful(tx_hash)

    def revoke_permission(self, did, address_to_revoke, account):
        """
        Revoke access permission to an address.

        :param did: the id of an asset on-chain, hex str
        :param address_to_revoke: ethereum account address, hex str
        :param account: Account executing the action
        :return: bool
        """
        tx_hash = self.send_transaction(
            'revokePermission',
            (did, address_to_revoke),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )
        return self.is_tx_successful(tx_hash)

    def get_permission(self, did, address):
        """
        Gets access permission of a grantee

        :param did: the id of an asset on-chain, hex str
        :param address: ethereum account address, hex str
        :return: true if the address has access permission to a DID
        """
        return self.contract.caller.getPermission(did, address)

    def is_did_provider(self, did, address):
        """
        Return true if the address is the Provider of the did.

        :param did: the id of an asset on-chain, hex str
        :param address: ethereum account address, hex str
        :return: bool
        """
        return self.contract.caller.isDIDProvider(did, address)

    def get_did_providers(self, did):
        """
        Return the list providers registered on-chain for the given did.

        :param did: hex str the id of an asset on-chain
        :return:
            list of addresses
            None if asset has no registerd providers
        """
        register_values = self.contract.caller.getDIDRegister(did)
        print(register_values)
        if register_values and len(register_values) >= 6 and register_values[0]:
            # sanitize providers list, because if providers were removed they will
            # be replaced with null/None
            valid_providers = [a for a in register_values[5] if
                               a != '0x0000000000000000000000000000000000000000']
            register_values[5] = valid_providers
            return DIDRegisterValues(*register_values).providers

    def get_owner_asset_ids(self, address):
        """
        Get the list of assets owned by an address owner.

        :param address: ethereum account address, hex str
        :return:
        """
        block_filter = self._get_event_filter(DIDRegistry.DID_REGISTRY_EVENT_NAME, owner=address)
        log_items = block_filter.get_all_entries(max_tries=5)
        did_list = []
        for log_i in log_items:
            did_list.append(log_i.args['_did'])

        return did_list

    def get_registered_attribute(self, did_bytes):
        """

        Example of event logs from event_filter.get_all_entries():
        [AttributeDict(
            {'args': AttributeDict(
                {'did': b'\x02n\xfc\xfb\xfdNM\xe9\xb8\xe0\xba\xc2\xb2\xc7\xbeg\xc9/\x95\xc3\x16\
                           x98G^\xb9\xe1\xf0T\xce\x83\xcf\xab',
                 'owner': '0xAd12CFbff2Cb3E558303334e7e6f0d25D5791fc2',
                 'value': 'http://localhost:5000',
                 'checksum': '0x...',
                 'updatedAt': 1947}),
             'event': 'DIDAttributeRegistered',
             'logIndex': 0,
             'transactionIndex': 1,
             'transactionHash': HexBytes(
             '0xea9ca5748d54766fb43fe9660dd04b2e3bb29a0fbe18414457cca3dd488d359d'),
             'address': '0x86DF95937ec3761588e6DEbAB6E3508e271cC4dc',
             'blockHash': HexBytes(
             '0xbbbe1046b737f33b2076cb0bb5ba85a840c836cf1ffe88891afd71193d677ba2'),
             'blockNumber': 1947})]

        """
        result = None
        did = Web3.toHex(did_bytes)
        block_number = self.get_block_number_updated(did_bytes)
        logger.debug(f'got blockNumber {block_number} for did {did}')
        if block_number == 0:
            raise DIDNotFound(
                f'DID "{did}" is not found on-chain in the current did registry. '
                f'Please ensure assets are registered in the correct keeper contracts. '
                f'The keeper-contracts DIDRegistry address is {self.address}.'
                f'Also ensure that sufficient time has passed after registering the asset '
                f'such that the transaction is confirmed by the network validators.')

        events = Web3Provider.get_web3().eth.getLogs({
            'fromBlock': block_number,
            'toBlock': block_number,
            'address': self.contract.address
        })
        event_data = self._filter_did_registered_events(did_bytes, events)

        if event_data:
            result = {
                'checksum': event_data.args._checksum,
                'value': event_data.args._value,
                'block_number': block_number,
                'did_bytes': event_data.args._did,
                'owner': Web3.toChecksumAddress(event_data.args._owner),
            }
        else:
            logger.warning(f'Could not find {DIDRegistry.DID_REGISTRY_EVENT_NAME} event logs for '
                           f'did {did} at blockNumber {block_number}')
        return result

    def _filter_did_registered_events(self, did_bytes, events):
        # https://ethereum.stackexchange.com/a/106981/3831

        event_template = getattr(self.contract.events, DIDRegistry.DID_REGISTRY_EVENT_NAME)
        for event in events:
            try:
                event_data  = get_event_data(event_template.web3.codec, event_template._get_event_abi(), event)
                if event_data['args']['_did'] == did_bytes:
                    return event_data
            except Exception:
                continue


    def _register_did(self, did, checksum, providers, url, account, activity_id=None,
                      attributes=None):
        assert isinstance(providers, list), ''
        return self.send_transaction(
            'registerDID',
            (did,
             checksum,
             providers,
             url,
             activity_id or '',
             attributes or ''),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )

    def _register_mintable_did(self, did, checksum, providers, url, account, cap, royalties, activity_id=None,
                               attributes=None):
        assert isinstance(providers, list), ''
        return self.send_transaction(
            'registerMintableDID',
            (did,
             checksum,
             providers,
             url,
             cap,
             royalties,
             activity_id or '',
             attributes or ''),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )

    def used(self, prov_id, did, agent_id, activity_id, signature, account, attributes=None):
        tx_hash = self.send_transaction(
            'used',
            (prov_id,
             did,
             agent_id,
             activity_id,
             signature,
             attributes),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )

        return Web3Provider.get_web3().eth.waitForTransactionReceipt(tx_hash, timeout=20)

    def was_derived_from(self, prov_id, new_entity_did, used_entity_did, agent_id, activity_id,
                         account, attributes=None):
        tx_hash = self.send_transaction(
            'wasDerivedFrom',
            (prov_id,
             new_entity_did,
             used_entity_did,
             agent_id,
             activity_id,
             attributes),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )

        return Web3Provider.get_web3().eth.waitForTransactionReceipt(tx_hash, timeout=20)

    def was_associated_with(self, prov_id, did, agent_id, activity_id,
                            account, attributes=None):
        tx_hash = self.send_transaction(
            'wasAssociatedWith',
            (prov_id,
             did,
             agent_id,
             activity_id,
             attributes),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )
        return Web3Provider.get_web3().eth.waitForTransactionReceipt(tx_hash, timeout=20)

    def acted_on_behalf(self, prov_id, did, delegate_agent_id, responsible_agent_id, activity_id,
                        signature, account, attributes=None):
        tx_hash = self.send_transaction(
            'actedOnBehalf',
            (prov_id,
             did,
             delegate_agent_id,
             responsible_agent_id,
             activity_id,
             signature,
             attributes),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )
        return Web3Provider.get_web3().eth.waitForTransactionReceipt(tx_hash, timeout=20)

    def add_did_provenance_delegate(self, did, delegate, account):
        tx_hash = self.send_transaction(
            'addDIDProvenanceDelegate',
            (did,
             delegate),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )
        return self.is_tx_successful(tx_hash)

    def remove_did_provenance_delegate(self, did, delegate, account):
        tx_hash = self.send_transaction(
            'removeDIDProvenanceDelegate',
            (did,
             delegate),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )
        return self.is_tx_successful(tx_hash)

    def is_provenance_delegate(self, did, delegate):
        return self.contract.caller.isProvenanceDelegate(did, delegate)

    def get_provenance_owner(self, did):
        return self.contract.caller.getProvenanceOwner(did)

    def is_provenance_signature_correct(self, agent_id, hash, signature):
        return self.contract.caller.provenanceSignatureIsCorrect(agent_id, hash, signature)

    def get_did_register(self, did):
        entry = self.contract.caller.getDIDRegister(did)
        valid_providers = [a for a in entry[5] if a != '0x0000000000000000000000000000000000000000']
        return {
            'owner': entry[0],
            'lastChecksum': entry[1],
            'url': entry[2],
            'lastUpdatedBy': entry[3],
            'blockNumberUpdated': entry[4],
            'providers': valid_providers,
            'nftSupply': entry[6],
            'mintCap': entry[7],
            'royalties': entry[8]
        }

    def get_provenance_entry(self, prov_id):
        provenance_entry = self.contract.caller.getProvenanceEntry(prov_id)
        return {'did': Web3.toHex(provenance_entry[0]),
                'related_did': Web3.toHex(provenance_entry[1]),
                'agent_id': provenance_entry[2],
                'activity_id': Web3.toHex(provenance_entry[3]),
                'agent_involved_id': provenance_entry[4],
                'method': provenance_entry[5],
                'created_by': provenance_entry[6],
                'block_number_updated': provenance_entry[7],
                'signature': provenance_entry[8]
                }

    def get_did_provenance_events(self, did_bytes):
        result = []
        did = Web3.toHex(did_bytes)
        block_filter = self._get_event_filter(
            DIDRegistry.PROVENANCE_ATTRIBUTE_REGISTERED_EVENT_NAME,
            did=did,
            from_block=0,
            to_block='latest')
        log_items = block_filter.get_all_entries(max_tries=5)
        if log_items:
            for i in range(len(log_items)):
                log_item = log_items[i].args
                block_number = log_item['_blockNumberUpdated']
                result.append({
                    'prov_id': log_item['provId'],
                    'did_bytes': log_item['_did'],
                    'agent_id': Web3.toChecksumAddress(log_item['_agentId']),
                    'activity_id': log_item['_activityId'],
                    'method': log_item['_method'],
                    'related_did': log_item['_relatedDid'],
                    'agent_involvedId': log_item['_agentInvolvedId'],
                    'attributes': log_item['_attributes'],
                    'block_number': block_number,

                })
        else:
            logger.warning(
                f'Could not find {DIDRegistry.PROVENANCE_ATTRIBUTE_REGISTERED_EVENT_NAME} event '
                f'logs for '
                f'did {did} at blockNumber ')
        return result

    def get_provenance_method_events(self, method, did_bytes):
        result = []
        did = Web3.toHex(did_bytes)
        try:
            event_name = self._method_mapper(method)
        except Exception as e:
            raise Exception(f'Provenance mapping exception: {e}')
        block_filter = self._get_event_filter(
            event_name,
            did=did
        )
        log_items = block_filter.get_all_entries(max_tries=5)
        if log_items:
            for i in range(len(log_items)):
                log_item = log_items[i].args
                block_number = log_item['_blockNumberUpdated']
                result.append({
                    'prov_id': log_item['provId'],
                    'did_bytes': did_bytes,
                    'agent_id': log_item.get('_agentId', None),
                    'activity_id': log_item['_activityId'],
                    'method': method,
                    'related_did': log_item.get('_relatedDid', None),
                    'agent_involvedId': log_item.get('_agentInvolvedId', None),
                    'attributes': log_item['_attributes'],
                    'block_number': block_number,

                })
        else:
            logger.warning(
                f'Could not find {DIDRegistry.PROVENANCE_ATTRIBUTE_REGISTERED_EVENT_NAME} event '
                f'logs for '
                f'did {did} at blockNumber')
        return result

    def mint(self, did, amount, account):
        tx_hash = self.send_transaction(
            'mint',
            (did, amount),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )

        return self.is_tx_successful(tx_hash)

    def burn(self, did, amount, account):
        tx_hash = self.send_transaction(
            'burn',
            (did, amount),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )
        return self.is_tx_successful(tx_hash)

    @staticmethod
    def is_nft_approved_for_all(account, operator):
        """
        Return true if the operator address is ERC1155 approved

        :param account: the account address to check, hex str
        :param operator: ethereum operator address, hex str
        :return: bool
        """
        nft = NFTUpgradeable.get_instance()
        return nft.is_approved_for_all(account, operator)

    def set_nft_proxy_approval(self, operator, approved, account):
        tx_hash = self.send_transaction(
            'setProxyApproval',
            (operator, approved),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )
        return self.is_tx_successful(tx_hash)


    def _get_event_filter(self, event_name, did=None, owner=None, from_block=0, to_block='latest'):
        _filters = {}
        if event_name == DIDRegistry.PROVENANCE_WAS_DERIVED_FROM_EVENT_NAME:
            _filters['_newEntityDid'] = Web3.toBytes(hexstr=did)
        elif event_name == DIDRegistry.PROVENANCE_WAS_ASSOCIATED_WITH_EVENT_NAME or event_name == DIDRegistry.PROVENANCE_ACTED_ON_BEHALF_EVENT_NAME:
            _filters['_entityDid'] = Web3.toBytes(hexstr=did)
        else:
            if did is not None:
                _filters['_did'] = Web3.toBytes(hexstr=did)
            if owner is not None:
                _filters['_owner'] = Web3.toBytes(hexstr=owner)

        block_filter = EventFilter(
            event_name,
            getattr(self.events, event_name),
            from_block=from_block,
            to_block=to_block,
            argument_filters=_filters,
        )
        return block_filter

    @staticmethod
    def _method_mapper(method):
        if method == 'ENTITY':
            raise ValueError(f'Method {method} not implemented')
        if method == 'ACTIVITY':
            raise ValueError(f'Method {method} not implemented')
        if method == 'WAS_GENERATED_BY':
            return DIDRegistry.PROVENANCE_WAS_GENERATED_BY_EVENT_NAME
        if method == 'USED':
            return DIDRegistry.PROVENANCE_USED_EVENT_NAME
        if method == 'WAS_INFORMED_BY':
            raise ValueError(f'Method {method} not implemented')
        if method == 'WAS_STARTED_BY':
            raise ValueError(f'Method {method} not implemented')
        if method == 'WAS_ENDED_BY':
            raise ValueError(f'Method {method} not implemented')
        if method == 'WAS_INVALIDATED_BY':
            raise ValueError(f'Method {method} not implemented')
        if method == 'WAS_DERIVED_FROM':
            return DIDRegistry.PROVENANCE_WAS_DERIVED_FROM_EVENT_NAME
        if method == 'AGENT':
            raise ValueError(f'Method {method} not implemented')
        if method == 'WAS_ATTRIBUTED_TO':
            raise ValueError(f'Method {method} not implemented')
        if method == 'WAS_ASSOCIATED_WITH':
            return DIDRegistry.PROVENANCE_WAS_ASSOCIATED_WITH_EVENT_NAME
        if method == 'ACTED_ON_BEHALF':
            return DIDRegistry.PROVENANCE_ACTED_ON_BEHALF_EVENT_NAME
        else:
            raise ValueError(f'Method {method} is not implemented in the provenance specificacion.')



import logging
from web3 import Web3
from contracts_lib_py import ContractBase
from contracts_lib_py.event_filter import EventFilter

logger = logging.getLogger(__name__)


class TemplateBase(ContractBase):
    """Class representing the AccessTemplate contract."""
    AGREEMENT_CREATED_EVENT = 'AgreementCreated'

    def create_agreement(self, agreement_id, did, condition_ids, time_locks, time_outs,
                         consumer_address, account):
        """
        Create the service agreement. Return true if it is created successfully.

        :param agreement_id: id of the agreement, hex str
        :param did: DID, str
        :param condition_ids: is a list of bytes32 content-addressed Condition IDs, bytes32
        :param time_locks: is a list of uint time lock values associated to each Condition, int
        :param time_outs: is a list of uint time out values associated to each Condition, int
        :param consumer_address: ethereum account address of consumer, hex str
        :param account: Account instance creating the agreement
        :return: bool
        """
        logger.debug(
            f'Creating agreement {agreement_id} with did={did}, consumer={consumer_address}.')
        tx_hash = self.send_transaction(
            'createAgreement',
            (agreement_id,
             did,
             condition_ids,
             time_locks,
             time_outs,
             consumer_address),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file},
        )
        return self.is_tx_successful(tx_hash)

    def get_condition_types(self):
        """

        :return:
        """
        return self.contract.caller.getConditionTypes()

    def get_agreement_data(self, agreement_id):
        """

        :param agreement_id: id of the agreement, hex str
        :return:
        """
        consumer, provider = self.contract.caller.getAgreementData(agreement_id)
        return consumer, provider

    def get_agreement_consumer(self, agreement_id):
        """

        :param agreement_id: id of the agreement, hex str
        :return:
        """
        data = self.get_agreement_data(agreement_id)
        return data[0] if data and len(data) > 1 else None

    def subscribe_agreement_created(self, agreement_id, timeout, callback, args, wait=False,
                                    from_block='latest', to_block='latest'):
        """
        Subscribe to an agreement created.

        :param agreement_id: id of the agreement, hex str
        :param timeout:
        :param callback:
        :param args:
        :param wait: if true block the listener until get the event, bool
        :param from_block: int or None
        :param to_block: int or None
        :return:
        """
        logger.debug(
            f'Subscribing {self.AGREEMENT_CREATED_EVENT} event with agreement id {agreement_id}.')
        return self.subscribe_to_event(
            self.AGREEMENT_CREATED_EVENT,
            timeout,
            {'_agreementId': Web3.toBytes(hexstr=agreement_id)},
            callback=callback,
            args=args,
            wait=wait,
            from_block=from_block,
            to_block=to_block
        )

    def get_event_filter_for_agreement_created(self, provider_address=None, from_block='latest',
                                               to_block='latest'):
        """

        :param provider_address: hex str ethereum address
        :param from_block: int or None
        :param to_block: int or None
        :return:
        """
        _filter = {}
        if provider_address:
            assert isinstance(provider_address, str)
            _filter['_accessProvider'] = provider_address

        event_filter = EventFilter(
            self.AGREEMENT_CREATED_EVENT,
            getattr(self.events, self.AGREEMENT_CREATED_EVENT),
            _filter,
            from_block=from_block,
            to_block=to_block
        )
        event_filter.set_poll_interval(0.5)
        return event_filter



import codecs

from eth_keys.datatypes import Signature
from eth_keys.utils.numeric import int_to_byte
from eth_keys.utils.padding import pad32
from eth_utils import int_to_big_endian


class SignatureFix(Signature):
    """
    Hack the Signature class to allow rebuilding of signature with a
    v value of 27 or 28 instead of 0 or 1
    """

    def __init__(self,
                 signature_bytes=None,
                 vrs=None,
                 backend=None,
                 ) -> None:
        v, r, s = vrs
        if v == 27 or v == 28:
            v -= 27

        vrs = (v, r, s)
        Signature.__init__(
            self,
            signature_bytes,
            vrs,
            backend,
        )

    def to_hex_v_hacked(self) -> str:
        # Need the 'type: ignore' comment below because of
        # https://github.com/python/typeshed/issues/300
        return '0x' + codecs.decode(codecs.encode(self.to_bytes_v_hacked(), 'hex'),
                                    'ascii')  # type: ignore

    def to_bytes_v_hacked(self) -> bytes:
        v = self.v
        if v == 0 or v == 1:
            v += 27
        vb = int_to_byte(v)
        rb = pad32(int_to_big_endian(self.r))
        sb = pad32(int_to_big_endian(self.s))
        # FIXME: Enable type checking once we have type annotations in eth_utils
        return b''.join((rb, sb, vb))  # type: ignore



"""Test NFTHolderCondition contract."""
from contracts_lib_py.conditions.transfer_did import TransferDIDOwnershipCondition


def test_transfer_did_condition_contract():
    transfer_did_condition = TransferDIDOwnershipCondition('TransferDIDOwnershipCondition')
    assert transfer_did_condition
    assert isinstance(transfer_did_condition, TransferDIDOwnershipCondition), \
        f'{transfer_did_condition} is not instance of TransferDIDOwnershipCondition'



from contracts_lib_py.conditions.condition_base import ConditionBase
from web3 import Web3


class LockPaymentCondition(ConditionBase):
    """Class representing the LockPaymentCondition contract."""
    CONTRACT_NAME = 'LockPaymentCondition'

    def fulfill(self, agreement_id, did, reward_address, token_address, amounts, receivers, account):
        """
        Fulfill the lock reward condition.

        :param agreement_id: id of the agreement, hex str
        :param did: id of the asset, hex str
        :param reward_address: the contract address where the reward is locked, hex str
        :param token_address: the token address to use. If empty  it will use 0x0 address that means ETH payment
        :param amounts: Array of amount of tokens to distribute, int[]
        :param receivers: Array of ethereum address of the receivers, hex str[]
        :param account: Account instance
        :return:
        """

        return self._fulfill(
            agreement_id,
            did,
            reward_address,
            self.validate_token_address(token_address),
            amounts,
            self.to_checksum_addresses(receivers),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'keyfile': account.key_file}
        )

    def hash_values(self, did, reward_address, token_address, amounts, receivers):
        """
        Hash of the values of the lock reward condition.

        :param did: id of the asset, hex str
        :param reward_address: the contract address where the reward is locked, hex str
        :param token_address: the token address to use. If empty  it will use 0x0 address that means ETH payment
        :param amounts: Array of amount of tokens to distribute, int[]
        :param receivers: Array of ethereum address of the receivers, hex str[]
        :return: hex str
        """

        return self._hash_values(
            did,
            reward_address,
            self.validate_token_address(token_address),
            amounts,
            self.to_checksum_addresses(receivers)
        )



from collections import namedtuple
from contracts_lib_py import ContractBase

AgreementTemplate = namedtuple(
    'AgreementTemplate',
    ('state', 'owner', 'updated_by', 'block_number_updated')
)


class TemplateStoreManager(ContractBase):
    """Class representing the TemplateStoreManager contract."""
    CONTRACT_NAME = 'TemplateStoreManager'

    def get_template(self, template_id):
        """
        Get the template for a given template id.

        :param template_id: id of the template, str
        :return:
        """
        template = self.contract.caller.getTemplate(template_id)
        if template and len(template) == 4:
            return AgreementTemplate(*template)

        return None

    def propose_template(self, template_id, from_account):
        """Propose a template.

        :param template_id: id of the template, str
        :param from_account: Account
        :return: bool
        """
        tx_hash = self.send_transaction(
            'proposeTemplate',
            (template_id,),
            transact={'from': from_account.address,
                      'passphrase': from_account.password,
                      'keyfile': from_account.key_file}
        )
        return self.is_tx_successful(tx_hash)

    def approve_template(self, template_id, from_account):
        """
        Approve a template.

        :param template_id: id of the template, str
        :param from_account: Account
        :return:
        """
        tx_hash = self.send_transaction(
            'approveTemplate',
            (template_id,),
            transact={'from': from_account.address,
                      'passphrase': from_account.password,
                      'keyfile': from_account.key_file}
        )
        return self.is_tx_successful(tx_hash)

    def revoke_template(self, template_id, from_account):
        """
        Revoke a template.

        :param template_id: id of the template, str
        :param from_account: Account
        :return: bool
        """
        tx_hash = self.send_transaction(
            'revokeTemplate',
            (template_id,),
            transact={'from': from_account.address,
                      'passphrase': from_account.password,
                      'keyfile': from_account.key_file}
        )
        return self.is_tx_successful(tx_hash)

    def is_template_approved(self, template_id):
        """
        True if the template is approved.

        :param template_id: id of the template, str
        :return: bool
        """
        return self.contract.caller.isTemplateApproved(template_id)

    def get_num_templates(self):
        """
        Return the number of templates on-chain.

        :return: number of templates, int
        """
        return self.contract.caller.getTemplateListSize()



from contracts_lib_py.contract_base import ContractBase


class GenericContract(ContractBase):
    """Class for instantiating any contract.

    Contract name is set at time of loading the contract.

    """


class GenericContractExternal(ContractBase):
    def __init__(self, address, abi, name, version='external'):
        from contracts_lib_py.contract_handler import ContractHandler

        self.name = name,
        self.CONTRACT_NAME = name
        self.version = version
        self.contract = ContractHandler.get_by_address(address, abi, name)


from collections import namedtuple

from eth_utils import add_0x_prefix

from contracts_lib_py import ContractBase

AgreementValues = namedtuple(
    'AgreementValues',
    ('did', 'owner', 'template_id', 'condition_ids', 'condition_id_seeds', 'id_seed')
)


class AgreementStoreManager(ContractBase):
    """Class representing the AgreementStoreManager contract."""
    CONTRACT_NAME = 'AgreementStoreManager'

    def create_agreement(self, agreement_id, did, condition_types, condition_ids, time_locks,
                         time_outs):
        """
        Create a new agreement.
        The agreement will create conditions of conditionType with conditionId.
        Only "approved" templates can access this function.
        :param agreement_id:id of the agreement, hex str
        :param did: DID of the asset. The DID must be registered beforehand, bytes32
        :param condition_types: is a list of addresses that point to Condition contracts,
                                list(address)
        :param condition_ids: is a list of bytes32 content-addressed Condition IDs, bytes32
        :param time_locks: is a list of uint time lock values associated to each Condition, int
        :param time_outs: is a list of uint time out values associated to each Condition, int
        :return: bool
        """

        tx_hash = self.contract.caller.createAgreement(
            agreement_id,
            did,
            condition_types,
            condition_ids,
            time_locks,
            time_outs,
        )
        return self.is_tx_successful(tx_hash)

    def set_templates(self, temp):
        self.templates = temp

    def get_agreement(self, agreement_id):
        """
        Retrieve the agreement for a agreement_id.

        :param agreement_id: id of the agreement, hex str
        :return: the agreement attributes.
        """
        # get info from events
        template = self.contract.caller.getAgreementTemplate(agreement_id)
        event = self.templates[template].subscribe_agreement_created(
            agreement_id, 15, None, (), wait=True, from_block=0
        )
        agreement = event.args
        did = add_0x_prefix(agreement._did.hex())
        cond_ids = [add_0x_prefix(_id.hex()) for _id in agreement._conditionIds]
        cond_id_seeds = [add_0x_prefix(_id.hex()) for _id in agreement._conditionIdSeeds]

        return AgreementValues(
            did,
            agreement._creator,
            template,
            cond_ids,
            cond_id_seeds,
            agreement._idSeed
        )

    def get_agreement_did_owner(self, agreement_id):
        """Get the DID owner for this agreement with _id.

        :param agreement_id: id of the agreement, hex str
        :return: the DID owner associated with agreement.did from the DID registry.
        """
        return self.contract.caller.getAgreementDIDOwner(agreement_id)

    def get_num_agreements(self):
        """Return the size of the Agreements list.

        :return: the length of the agreement list, int
        """
        return self.contract.caller.getAgreementListSize()

    def hash_id(self, did_seed, address):
        """
        Calculates the final DID given a DID seed and the publisher address
        :param did_seed: the hash used to generate the did
        :param address: the address of the account creating the asset
        :return: the final DID
        """
        return add_0x_prefix(self.contract.caller.agreementId(did_seed, address).hex())



"""Test AccessConditions contract."""
import uuid

from contracts_lib_py.conditions.access import AccessCondition
from contracts_lib_py.conditions.access_proof import AccessProofCondition
from contracts_lib_py.conditions.nft_access import NFTAccessCondition
from tests.resources.helper_functions import get_consumer_account


def test_access_condition_contract():
    access_condition = AccessCondition('AccessCondition')
    assert access_condition
    assert isinstance(access_condition, AccessCondition), \
        f'{access_condition} is not instance of AccessCondition'

def test_access_proof_condition_contract():
    access_proof_condition = AccessProofCondition('AccessProofCondition')
    assert access_proof_condition
    assert isinstance(access_proof_condition, AccessProofCondition), \
        f'{access_proof_condition} is not instance of AccessProofCondition'

def test_nft_access_condition_contract():
    nft_access_condition = NFTAccessCondition('NFTAccessCondition')
    assert nft_access_condition
    assert isinstance(nft_access_condition, NFTAccessCondition), \
        f'{nft_access_condition} is not instance of NFTAccessCondition'


def test_check_permissions_not_registered_did():
    access_condition = AccessCondition('AccessCondition')
    consumer_account = get_consumer_account()
    did_id = uuid.uuid4().hex + uuid.uuid4().hex
    assert not access_condition.check_permissions(did_id, consumer_account.address)

# TODO Create test for check permission after access granted.



import logging

from contracts_lib_py.templates.template_base import TemplateBase

logger = logging.getLogger('nftAccessTemplate')


class NFTAccessTemplate(TemplateBase):
    """Class representing the NFTAccessTemplate contract."""
    CONTRACT_NAME = 'NFTAccessTemplate'



import os

from eth_utils import add_0x_prefix
from web3 import Web3

from contracts_lib_py import Keeper
from contracts_lib_py.contract_handler import ContractHandler
from contracts_lib_py.utils import prepare_prefixed_hash, add_ethereum_prefix_and_hash_msg
from tests.resources.helper_functions import get_publisher_account, get_resource_path


def test_keeper_instance():
    keeper = Keeper(ContractHandler.artifacts_path)
    assert keeper
    assert isinstance(keeper.get_instance(), Keeper)


def test_ec_recover():
    test_values = [
        ('0xe2DD09d719Da89e5a3D0F2549c7E24566e947260',
         'c80996119e884cb38599bcd96a22ad3eea3a4734bcfb47959a5d41ecdcbdfe67',
         '0xa50427a9d5beccdea3eeabecfc1014096b35cd05965e772e8ea32477d2f217'
         'c30d0ec5dbf6b14de1d6eeff45011d17490fe5126576b20d2cbada828cb068c9f801'),
        ('0x00Bd138aBD70e2F00903268F3Db08f2D25677C9e',
         'd77c3a84cafe4cb8bc28bf41a99c63fd530c10da33a54acf94e8d1369d09fbb2',
         '0x9076b561e554cf657af333d9680ba118d556c5b697622636bce4b02f4d5632'
         '5a0ea6a474ca85291252c8c1b8637174ee32072bef357bb0c21b0db4c25b379e781b'),
        ('0xe2DD09d719Da89e5a3D0F2549c7E24566e947260',
         '8d5c1065a9c74da59fbb9e41d1f196e40517e92d81b14c3a8143d6887f3f4438',
         '0x662f6cffd96ada4b6ce5497d444c92126bd053ab131915332edf0dbba716ba'
         '82662275670c95eb2a4d65245cac70313c25e34f594d7c0fbca5232c3d5701a57e00')
    ]

    for expected_address, document_id, signed_document_id in test_values:
        rec_address = Keeper.get_instance().ec_recover(document_id, signed_document_id)
        if expected_address.lower() == rec_address.lower():
            msg_hsh = prepare_prefixed_hash(document_id)
            rec_address = Keeper.get_instance().ec_recover(msg_hsh, signed_document_id)
            assert expected_address.lower() == rec_address.lower()


def test_artifacts_path():
    artifacts_path = '/some/other/path'
    keeper = Keeper.get_instance(artifacts_path)
    assert keeper.artifacts_path == artifacts_path


def test_get_network_id():
    network_id = Keeper.get_network_id()
    assert isinstance(network_id, int)
    assert network_id in Keeper._network_name_map


def test_get_network_name():
    name = Keeper.get_network_name(Keeper.get_network_id())
    assert name in Keeper._network_name_map.values()
    os.environ['KEEPER_NETWORK_NAME'] = 'yellow'
    assert 'KEEPER_NETWORK_NAME' in os.environ
    name = Keeper.get_network_name(Keeper.get_network_id())
    assert name == 'yellow'
    del os.environ['KEEPER_NETWORK_NAME']

    assert Keeper.get_network_name(1) == Keeper._network_name_map.get(1)
    assert Keeper.get_network_name(2) == Keeper._network_name_map.get(2)
    assert Keeper.get_network_name(3) == Keeper._network_name_map.get(3)
    assert Keeper.get_network_name(4) == Keeper._network_name_map.get(4)
    assert Keeper.get_network_name(42) == Keeper._network_name_map.get(42)
    assert Keeper.get_network_name(77) == Keeper._network_name_map.get(77)
    assert Keeper.get_network_name(99) == Keeper._network_name_map.get(99)
    assert Keeper.get_network_name(137) == Keeper._network_name_map.get(137)
    assert Keeper.get_network_name(8996) == Keeper._network_name_map.get(8996)
    assert Keeper.get_network_name(8997) == Keeper._network_name_map.get(8997)
    assert Keeper.get_network_name(42220) == Keeper._network_name_map.get(42220)
    assert Keeper.get_network_name(44787) == Keeper._network_name_map.get(44787)
    assert Keeper.get_network_name(62320) == Keeper._network_name_map.get(62320)
    assert Keeper.get_network_name(80001) == Keeper._network_name_map.get(80001)
    assert Keeper.get_network_name(1313161554) == Keeper._network_name_map.get(1313161554)
    assert Keeper.get_network_name(1313161555) == Keeper._network_name_map.get(1313161555)
    assert Keeper.get_network_name(1313161556) == Keeper._network_name_map.get(1313161556)
    assert Keeper.get_network_name(0) == 'development'


def test_verify_signature(web3_instance):
    """
    `keeper` lib currently uses `web3.eth.sign()` to sign the service agreement
    hash. This signing method uses ethereum `eth_sign` on the ethereum client
    which automatically prepends the message with text defined in EIP-191 as
    version 'E': `b'\\x19Ethereum Signed Message:\\n'` concatenated with the
    number of bytes in the message.

    It is more convenient to sign a message using `web3.eth.sign()` because it only requires the
    account address
    whereas `web3.eth.account.sign_hash()` requires a private_key to sign the message.
    `web3.eth.account.sign_hash()` also does not prepend anything to the message before signing.
    Messages signed via Metamask in pleuston use the latter method and currently fail to verify in
    `keeper`.
    The signature verification fails because recoverHash is being used on a prepended message but
    the signature created by `web3.eth.account.sign_hash()` does not add a prefix before signing.
    """
    w3 = web3_instance

    # Signature created from Metamask (same as using `web3.eth.account.sign_hash()`)
    address = '0x8248039e67801Ac0B9d0e38201E963194abdb540'
    hex_agr_hash = '0xc8ea6bf6f4f4e2bf26a645dd4a1be20f5151c74964026c36efc2149bfae5f924'
    agreement_hash = Web3.toBytes(hexstr=hex_agr_hash)
    assert hex_agr_hash == '0x' + agreement_hash.hex()
    signature = (
        '0x200ce6aa55f0b4080c5f3a5dbe8385d2d196b0380cbdf388f79b6b004223c68a4f7972deb36417df8599155da2f903e43fe7e7eb40214db6bd6e55fd4c4fcf2a1c'
    )
    recovered_address = w3.eth.account.recoverHash(agreement_hash, signature=signature)
    assert address == recovered_address, f'Could not verify signature using address {address}'

    # Signature created using `web3.eth.sign()` (squid-py, squid-js with no metamask)
    address = "0x00Bd138aBD70e2F00903268F3Db08f2D25677C9e"
    hex_agr_hash = "0xeeaae0098b39fdf8fab6733152dd0ef54729ac486f9846450780c5cc9d44f5e8"
    agreement_hash = Web3.toBytes(hexstr=hex_agr_hash)
    signature = (
        "0x44fa549d33f5993f73e96f91cad01d9b37830da78494e35bda32a280d1b864ac020a761e872633c8149a5b63b65a1143f9f5a3be35822a9e90e0187d4a1f9d101c"
    )
    assert hex_agr_hash == add_0x_prefix(agreement_hash.hex())
    prefixed_hash = prepare_prefixed_hash(agreement_hash)
    recovered_address = w3.eth.account.recoverHash(prefixed_hash, signature=signature)
    assert address == recovered_address, f'Could not verify signature using address {address}'


def test_sign_and_recover(web3_instance):
    w3 = web3_instance
    account = get_publisher_account()
    msg = 'testing-signature-and-recovery-of-signer-address'
    msg_hash = w3.keccak(text=msg)
    signature = Keeper.sign_hash(msg_hash, account)
    address = w3.toChecksumAddress(Keeper.ec_recover(msg_hash, signature))
    assert address == account.address

    # Signature created on msg with the ethereum prefix. `web3.eth.account.recoverHash` does NOT
    # add any prefixes to the message, so we have to add the prefix before the call.
    address = w3.eth.account.recoverHash(msg_hash, signature=signature)
    assert address == account.address

    # Now do the opposite, sign with eth.account.sign_hash() (using prefixed msg hash),
    # then recover address with Keeper.ec_recover() on the msg hash with no prefix.
    with open(get_resource_path('data', 'publisher_key_file.json')) as kf:
        key = kf.read()
    prvkey = w3.eth.account.decrypt(key, account.password)
    account_sig_prefixed = add_0x_prefix(w3.eth.account.signHash(msg_hash, prvkey)['signature'].hex())
    assert Keeper.ec_recover(msg_hash, account_sig_prefixed).lower() == account.address.lower()

    # Test specific case where message is signed by some Wallet web3 such as Metamask or
    # burner wallet. Such signature uses the `web3.personal` `sign` method which adds
    # `Ethereum Signed Message` prefix in a generic way, see `add_ethereum_prefix_and_hash_msg` for details.
    sig = '0xa9e78d2c088c0b17a8c35b69e0dfa774692ccabed570e40502795bd41f561cf7677ed02bf4ee7967a55979d585bbf203b4a490e1d747e5a4d60a50859d816ac51b'
    publisher_address = '0x903322C7E45A60d7c8C3EA236c5beA9Af86310c7'
    doc_id = '028faa498d154388a89dc0dea908a4e27700920217a44abe8f1cdd64953125b8'
    prefixed_hash = add_ethereum_prefix_and_hash_msg(doc_id)
    recovered_address = w3.eth.account.recoverHash(prefixed_hash, signature=sig)
    assert recovered_address == publisher_address
    recovered_address = Keeper.ec_recover(prefixed_hash, sig)
    assert recovered_address == publisher_address

    recovered_address = Keeper.personal_ec_recover(doc_id, sig)
    assert recovered_address == publisher_address


def test_load_external_contract(external_contract):
    address, abi, name = external_contract
    keeper = Keeper.get_instance(external_contracts=[external_contract])

    contract = keeper.get_contract(name)
    assert contract is not None
    assert contract.address == address
    assert contract.CONTRACT_NAME == name


import os

import pytest
from web3 import Web3
import nevermined_contracts

from contracts_lib_py.web3_provider import Web3Provider
from contracts_lib_py.web3.http_provider import CustomHTTPProvider
from contracts_lib_py.contract_handler import ContractHandler
from contracts_lib_py.keeper import Keeper

INFURA_TOKEN = os.environ.get("INFURA_TOKEN")

pytestmark = pytest.mark.integration


@pytest.mark.parametrize("keeper_url,network_id,network_name", [
    [f"https://mainnet.infura.io/v3/{INFURA_TOKEN}", [1], ["mainnet"]],
    [f"https://rinkeby.infura.io/v3/{INFURA_TOKEN}", [4], ["rinkeby"]],
    ("http://localhost:8545", [8996, 8997], ["spree", "polygon-localnet"]),
    (f"https://polygon-mumbai.infura.io/v3/{INFURA_TOKEN}", [80001], ["mumbai"]),
    ("https://alfajores-forno.celo-testnet.org", [44787], ["celo-alfajores"]),
    (f"https://polygon-mainnet.infura.io/v3/{INFURA_TOKEN}", [137], ["matic"]),
    ("https://forno.celo.org", [42220], ["celo"])
# Baklava problems...
#    ("https://baklava-forno.celo-testnet.org", [62320], ["celo-baklava"])
    ],
    ids=[
        "mainnet",
        "rinkeby",
        "spree/polygon-localnet",
        "mumbai",
        "celo-alfajores",
        "matic",
        "celo"
#        "celo-baklava"
    ]
)
def test_artifacts(keeper_url, network_id, network_name):
    Web3Provider._web3 = Web3(CustomHTTPProvider(keeper_url))
    ContractHandler.artifacts_path = nevermined_contracts.get_artifacts_path()
    keeper = Keeper()
    assert keeper.get_network_id() in network_id
    assert keeper.get_network_name(keeper.get_network_id()) in network_name
    assert keeper.did_registry is not None
